#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#


from airbyte_cdk.sources.file_based.exceptions import ConfigValidationError, FileBasedSourceError
from unit_tests.sources.file_based.scenarios.scenario_builder import TestScenarioBuilder

"""
User input schema rules:
  - `check`: Successful if the schema conforms to a record, otherwise ConfigValidationError.
  - `discover`: User-input schema is output if the schema is valid, otherwise ConfigValidationError.
  - `read`: If the schema is valid, record values are cast to types in the schema; if this is successful
    the records are emitted. otherwise an error is logged. If the schema is not valid, ConfigValidationError.
"""


_base_user_input_schema_scenario = (
    TestScenarioBuilder()
    .set_files(
        {
            "a.csv": {
                "contents": [
                    ("col1", "col2"),
                    ("val11", "val12"),
                    ("val21", "val22"),
                ],
                "last_modified": "2023-06-05T03:54:07.000Z",
            }
        }
    )
    .set_file_type("csv")
    .set_expected_catalog(
        {
            "streams": [
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col1": {
                                "type": "string"
                            },
                            "col2": {
                                "type": "string"
                            },
                            "_ab_source_file_last_modified": {
                                "type": "string"
                            },
                            "_ab_source_file_url": {
                                "type": "string"
                            },
                        },
                    },
                    "name": "stream1",
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                }
            ]
        }
    )
    .set_expected_records(
        [
            {"data": {"col1": "val11", "col2": "val12", "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                      "_ab_source_file_url": "a.csv"}, "stream": "stream1"},
            {"data": {"col1": "val21", "col2": "val22", "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                      "_ab_source_file_url": "a.csv"}, "stream": "stream1"},
        ]
    )
)


valid_single_stream_user_input_schema_scenario = (
    _base_user_input_schema_scenario.copy()
    .set_name("valid_single_stream_user_input_schema_scenario")
    .set_config(
        {
            "streams": [
                {
                    "name": "stream1",
                    "file_type": "csv",
                    "globs": ["*"],
                    "validation_policy": "emit_record",
                    "input_schema": '{"col1": "string", "col2": "string"}',
                }
            ]
        }
    )
    .set_expected_check_status("SUCCEEDED")
).build()


single_stream_user_input_schema_scenario_schema_is_invalid = (
    _base_user_input_schema_scenario.copy()
    .set_name("single_stream_user_input_schema_scenario_schema_is_invalid")
    .set_config(
        {
            "streams": [
                {
                    "name": "stream1",
                    "file_type": "csv",
                    "globs": ["*"],
                    "validation_policy": "emit_record",
                    "input_schema": '{"col1": "x", "col2": "string"}',
                }
            ]
        }
    )
    .set_expected_check_status("FAILED")
    .set_expected_check_error(None, FileBasedSourceError.ERROR_PARSING_USER_PROVIDED_SCHEMA.value)
    .set_expected_discover_error(ConfigValidationError, FileBasedSourceError.ERROR_PARSING_USER_PROVIDED_SCHEMA.value)
    .set_expected_read_error(ConfigValidationError, FileBasedSourceError.ERROR_PARSING_USER_PROVIDED_SCHEMA.value)
).build()


single_stream_user_input_schema_scenario_emit_nonconforming_records = (
    _base_user_input_schema_scenario.copy()
    .set_name("single_stream_user_input_schema_scenario_emit_nonconforming_records")
    .set_config(
        {
            "streams": [
                {
                    "name": "stream1",
                    "file_type": "csv",
                    "globs": ["*"],
                    "validation_policy": "emit_record",
                    "input_schema": '{"col1": "integer", "col2": "string"}',
                }
            ]
        }
    )
    .set_expected_check_status("FAILED")
    .set_expected_check_error(None, FileBasedSourceError.SCHEMA_INFERENCE_ERROR.value)
    .set_expected_catalog(
        {
            "streams": [
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col1": {
                                "type": "integer"
                            },
                            "col2": {
                                "type": "string"
                            },
                            "_ab_source_file_last_modified": {
                                "type": "string"
                            },
                            "_ab_source_file_url": {
                                "type": "string"
                            },
                        },
                    },
                    "name": "stream1",
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                }
            ]
        }
    )
).build()


single_stream_user_input_schema_scenario_skip_nonconforming_records = (
    _base_user_input_schema_scenario.copy()
    .set_name("single_stream_user_input_schema_scenario_skip_nonconforming_records")
    .set_config(
        {
            "streams": [
                {
                    "name": "stream1",
                    "file_type": "csv",
                    "globs": ["*"],
                    "validation_policy": "skip_record",
                    "input_schema": '{"col1": "integer", "col2": "string"}',
                }
            ]
        }
    )
    .set_expected_check_status("FAILED")
    .set_expected_check_error(None, FileBasedSourceError.SCHEMA_INFERENCE_ERROR.value)
    .set_expected_catalog(
        {
            "streams": [
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col1": {
                                "type": "integer"
                            },
                            "col2": {
                                "type": "string"
                            },
                            "_ab_source_file_last_modified": {
                                "type": "string"
                            },
                            "_ab_source_file_url": {
                                "type": "string"
                            },
                        },
                    },
                    "name": "stream1",
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                }
            ]
        }
    )
    .set_expected_records([])
    .set_expected_logs({
        "read": [
            {
                'level': 'WARN',
                'message': 'Records in file did not pass validation policy. stream=stream1 file=a.csv n_skipped=2 validation_policy=skip_record',
            },
            {
                'level': 'WARNING',
                'message': 'Could not cast the value to the expected type.: col1: value=val11,expected_type=integer',
            },
            {
                'level': 'WARNING',
                'message': 'Could not cast the value to the expected type.: col1: value=val21,expected_type=integer',
            },
        ]
    })
).build()


_base_multi_stream_user_input_schema_scenario = (
    TestScenarioBuilder()
    .set_files(
        {
            "a.csv": {
                "contents": [
                    ("col1", "col2"),
                    ("val11a", 21),
                    ("val12a", 22),
                ],
                "last_modified": "2023-06-05T03:54:07.000Z",
            },
            "b.csv": {
                "contents": [
                    ("col1", "col2", "col3"),
                    ("val11b", "val12b", "val13b"),
                    ("val21b", "val22b", "val23b"),
                ],
                "last_modified": "2023-06-05T03:54:07.000Z",
            },
            "c.csv": {
                "contents": [
                    ("col1",),
                    ("val11c",),
                    ("val21c",),
                ],
                "last_modified": "2023-06-05T03:54:07.000Z",
            },
        }
    )
    .set_file_type("csv")
    .set_expected_catalog(
        {
            "streams": [
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col1": {
                                "type": "string",
                            },
                            "col2": {
                                "type": "integer",
                            },
                            "_ab_source_file_last_modified": {
                                "type": "string"
                            },
                            "_ab_source_file_url": {
                                "type": "string"
                            },
                        }
                    },
                    "name": "stream1",
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                },
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col1": {
                                "type": "string",
                            },
                            "col2": {
                                "type": "string",
                            },
                            "col3": {
                                "type": "string",
                            },
                            "_ab_source_file_last_modified": {
                                "type": "string"
                            },
                            "_ab_source_file_url": {
                                "type": "string"
                            },
                        }
                    },
                    "name": "stream2",
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                },
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col1": {
                                "type": ["null", "string"],
                            },
                            "_ab_source_file_last_modified": {
                                "type": "string"
                            },
                            "_ab_source_file_url": {
                                "type": "string"
                            },
                        }
                    },
                    "name": "stream3",
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                },
            ]
        }
    )
    .set_expected_records(
        [
            {"data": {"col1": "val11a", "col2": 21, "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                      "_ab_source_file_url": "a.csv"}, "stream": "stream1"},
            {"data": {"col1": "val12a", "col2": 22, "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                      "_ab_source_file_url": "a.csv"}, "stream": "stream1"},
            # The files in b.csv are emitted despite having an invalid schema
            {"data": {"col1": "val11b", "col2": "val12b", "col3": "val13b", "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                      "_ab_source_file_url": "b.csv"}, "stream": "stream2"},
            {"data": {"col1": "val21b", "col2": "val22b", "col3": "val23b", "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                      "_ab_source_file_url": "b.csv"}, "stream": "stream2"},
            {"data": {"col1": "val11c", "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                      "_ab_source_file_url": "c.csv"}, "stream": "stream3"},
            {"data": {"col1": "val21c", "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                      "_ab_source_file_url": "c.csv"}, "stream": "stream3"},
        ]
    )
)


valid_multi_stream_user_input_schema_scenario = (
    _base_multi_stream_user_input_schema_scenario.copy()
    .set_name("valid_multi_stream_user_input_schema_scenario")
    .set_config(
        {
            "streams": [
                {
                    "name": "stream1",
                    "file_type": "csv",
                    "globs": ["a.csv"],
                    "validation_policy": "emit_record",
                    "input_schema": '{"col1": "string", "col2": "integer"}',
                },
                {
                    "name": "stream2",
                    "file_type": "csv",
                    "globs": ["b.csv"],
                    "validation_policy": "emit_record",
                    "input_schema": '{"col1": "string", "col2": "string", "col3": "string"}',
                },
                {
                    "name": "stream3",
                    "file_type": "csv",
                    "globs": ["c.csv"],
                    "validation_policy": "emit_record",
                },

            ]
        }
    )
    .set_expected_check_status("SUCCEEDED")
).build()


multi_stream_user_input_schema_scenario_schema_is_invalid = (
    _base_multi_stream_user_input_schema_scenario.copy()
    .set_name("multi_stream_user_input_schema_scenario_schema_is_invalid")
    .set_config(
        {
            "streams": [
                {
                    "name": "stream1",
                    "file_type": "csv",
                    "globs": ["a.csv"],
                    "validation_policy": "emit_record",
                    "input_schema": '{"col1": "string", "col2": "integer"}',
                },
                {
                    "name": "stream2",
                    "file_type": "csv",
                    "globs": ["b.csv"],
                    "validation_policy": "emit_record",
                    "input_schema": '{"col1": "x", "col2": "string", "col3": "string"}',  # this stream's schema is invalid
                },
                {
                    "name": "stream3",
                    "file_type": "csv",
                    "globs": ["c.csv"],
                    "validation_policy": "emit_record",
                },

            ]
        }
    )
    .set_expected_check_status("FAILED")
    .set_expected_check_error(None, FileBasedSourceError.ERROR_PARSING_USER_PROVIDED_SCHEMA.value)
    .set_expected_discover_error(ConfigValidationError, FileBasedSourceError.ERROR_PARSING_USER_PROVIDED_SCHEMA.value)
    .set_expected_read_error(ConfigValidationError, FileBasedSourceError.ERROR_PARSING_USER_PROVIDED_SCHEMA.value)
).build()


multi_stream_user_input_schema_scenario_emit_nonconforming_records = (
    _base_multi_stream_user_input_schema_scenario.copy()
    .set_name("multi_stream_user_input_schema_scenario_emit_nonconforming_records")
    .set_config(
        {
            "streams": [
                {
                    "name": "stream1",
                    "file_type": "csv",
                    "globs": ["a.csv"],
                    "validation_policy": "emit_record",
                    "input_schema": '{"col1": "string", "col2": "integer"}',
                },
                {
                    "name": "stream2",
                    "file_type": "csv",
                    "globs": ["b.csv"],
                    "validation_policy": "emit_record",
                    "input_schema": '{"col1": "string", "col2": "integer", "col3": "string"}',  # this stream's records do not conform to the schema
                },
                {
                    "name": "stream3",
                    "file_type": "csv",
                    "globs": ["c.csv"],
                    "validation_policy": "emit_record",
                },

            ]
        }
    )
    .set_expected_catalog(
        {
            "streams": [
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col1": {
                                "type": "string"
                            },
                            "col2": {
                                "type": "integer"
                            },
                            "_ab_source_file_last_modified": {
                                "type": "string"
                            },
                            "_ab_source_file_url": {
                                "type": "string"
                            },
                        },
                    },
                    "name": "stream1",
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                },
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col1": {
                                "type": "string"
                            },
                            "col2": {
                                "type": "integer"
                            },
                            "col3": {
                                "type": "string"
                            },
                            "_ab_source_file_last_modified": {
                                "type": "string"
                            },
                            "_ab_source_file_url": {
                                "type": "string"
                            },
                        },
                    },
                    "name": "stream2",
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                },
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col1": {
                                "type": ["null", "string"],
                            },
                            "_ab_source_file_last_modified": {
                                "type": "string"
                            },
                            "_ab_source_file_url": {
                                "type": "string"
                            },
                        },
                    },
                    "name": "stream3",
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                },
            ]
        }
    )
    .set_expected_check_status("FAILED")
    .set_expected_check_error(None, FileBasedSourceError.ERROR_PARSING_USER_PROVIDED_SCHEMA.value)
    .set_expected_records(
        [
            {"data": {"col1": "val11a", "col2": 21, "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                      "_ab_source_file_url": "a.csv"}, "stream": "stream1"},
            {"data": {"col1": "val12a", "col2": 22, "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                      "_ab_source_file_url": "a.csv"}, "stream": "stream1"},
            {"data": {"col1": "val11b", "col2": "val12b", "col3": "val13b", "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                      "_ab_source_file_url": "b.csv"}, "stream": "stream2"},
            {"data": {"col1": "val21b", "col2": "val22b", "col3": "val23b", "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                      "_ab_source_file_url": "b.csv"}, "stream": "stream2"},
            {"data": {"col1": "val11c", "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                      "_ab_source_file_url": "c.csv"}, "stream": "stream3"},
            {"data": {"col1": "val21c", "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                      "_ab_source_file_url": "c.csv"}, "stream": "stream3"},
        ]
    )
    .set_expected_logs({
        "read": [
            {
                'level': 'WARNING',
                'message': 'Could not cast the value to the expected type.: col2: value=val12b,expected_type=integer',
            },
            {
                'level': 'WARNING',
                'message': 'Could not cast the value to the expected type.: col2: value=val22b,expected_type=integer',
            },
        ]

    })
).build()


multi_stream_user_input_schema_scenario_skip_nonconforming_records = (
    _base_multi_stream_user_input_schema_scenario.copy()
    .set_name("multi_stream_user_input_schema_scenario_skip_nonconforming_records")
    .set_config(
        {
            "streams": [
                {
                    "name": "stream1",
                    "file_type": "csv",
                    "globs": ["a.csv"],
                    "validation_policy": "emit_record",
                    "input_schema": '{"col1": "string", "col2": "integer"}',
                },
                {
                    "name": "stream2",
                    "file_type": "csv",
                    "globs": ["b.csv"],
                    "validation_policy": "skip_record",
                    "input_schema": '{"col1": "string", "col2": "integer", "col3": "string"}',  # this stream's records do not conform to the schema
                },
                {
                    "name": "stream3",
                    "file_type": "csv",
                    "globs": ["c.csv"],
                    "validation_policy": "emit_record",
                },

            ]
        }
    )
    .set_expected_catalog(
        {
            "streams": [
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col1": {
                                "type": "string"
                            },
                            "col2": {
                                "type": "integer"
                            },
                            "_ab_source_file_last_modified": {
                                "type": "string"
                            },
                            "_ab_source_file_url": {
                                "type": "string"
                            },
                        },
                    },
                    "name": "stream1",
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                },
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col1": {
                                "type": "string"
                            },
                            "col2": {
                                "type": "integer"
                            },
                            "col3": {
                                "type": "string"
                            },
                            "_ab_source_file_last_modified": {
                                "type": "string"
                            },
                            "_ab_source_file_url": {
                                "type": "string"
                            },
                        },
                    },
                    "name": "stream2",
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                },
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col1": {
                                "type": ["null", "string"],
                            },
                            "_ab_source_file_last_modified": {
                                "type": "string"
                            },
                            "_ab_source_file_url": {
                                "type": "string"
                            },
                        },
                    },
                    "name": "stream3",
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                },
            ]
        }
    )
    .set_expected_check_status("FAILED")
    .set_expected_check_error(None, FileBasedSourceError.ERROR_PARSING_USER_PROVIDED_SCHEMA.value)
    .set_expected_records(
        [
            {"data": {"col1": "val11a", "col2": 21, "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                      "_ab_source_file_url": "a.csv"}, "stream": "stream1"},
            {"data": {"col1": "val12a", "col2": 22, "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                      "_ab_source_file_url": "a.csv"}, "stream": "stream1"},
            # {"data": {"col1": "val11b", "col2": "val12b", "col3": "val13b", "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
            #           "_ab_source_file_url": "b.csv"}, "stream": "stream2"},
            # {"data": {"col1": "val21b", "col2": "val22b", "col3": "val23b", "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
            #           "_ab_source_file_url": "b.csv"}, "stream": "stream2"},
            {"data": {"col1": "val11c", "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                      "_ab_source_file_url": "c.csv"}, "stream": "stream3"},
            {"data": {"col1": "val21c", "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                      "_ab_source_file_url": "c.csv"}, "stream": "stream3"},
        ]
    )
    .set_expected_logs({
        "read": [
            {
                'level': 'WARN',
                'message': 'Records in file did not pass validation policy. stream=stream2 file=b.csv n_skipped=2 validation_policy=skip_record',
            },
            {
                'level': 'WARNING',
                'message': 'Could not cast the value to the expected type.: col2: value=val12b,expected_type=integer',
            },
            {
                'level': 'WARNING',
                'message': 'Could not cast the value to the expected type.: col2: value=val22b,expected_type=integer',
            },
        ]
    })
).build()



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#


from unittest.mock import MagicMock, patch

import pytest
from airbyte_cdk.sources.declarative.requesters.error_handlers.backoff_strategies.wait_until_time_from_header_backoff_strategy import (
    WaitUntilTimeFromHeaderBackoffStrategy,
)

SOME_BACKOFF_TIME = 60
REGEX = "[-+]?\\d+"


@pytest.mark.parametrize(
    "test_name, header, wait_until, min_wait, regex, expected_backoff_time",
    [
        ("test_wait_until_time_from_header", "wait_until", 1600000060.0, None, None, 60),
        ("test_wait_until_time_from_header_parameters", "{{parameters['wait_until']}}", 1600000060.0, None, None, 60),
        ("test_wait_until_time_from_header_config", "{{config['wait_until']}}", 1600000060.0, None, None, 60),
        ("test_wait_until_negative_time", "wait_until", 1500000000.0, None, None, None),
        ("test_wait_until_time_less_than_min", "wait_until", 1600000060.0, 120, None, 120),
        ("test_wait_until_no_header", "absent_header", 1600000000.0, None, None, None),
        ("test_wait_until_time_from_header_not_numeric", "wait_until", "1600000000,1600000000", None, None, None),
        ("test_wait_until_time_from_header_is_numeric", "wait_until", "1600000060", None, None, 60),
        ("test_wait_until_time_from_header_with_regex", "wait_until", "1600000060,60", None, "[-+]?\d+", 60),  # noqa
        ("test_wait_until_time_from_header_with_regex_from_parameters", "wait_until", "1600000060,60", None, "{{parameters['regex']}}", 60),
        # noqa
        ("test_wait_until_time_from_header_with_regex_from_config", "wait_until", "1600000060,60", None, "{{config['regex']}}", 60),  # noqa
        ("test_wait_until_time_from_header_with_regex_no_match", "wait_time", "...", None, "[-+]?\d+", None),  # noqa
        ("test_wait_until_no_header_with_min", "absent_header", "1600000000.0", SOME_BACKOFF_TIME, None, SOME_BACKOFF_TIME),
        (
            "test_wait_until_no_header_with_min_from_parameters",
            "absent_header",
            "1600000000.0",
            "{{parameters['min_wait']}}",
            None,
            SOME_BACKOFF_TIME,
        ),
        (
            "test_wait_until_no_header_with_min_from_config",
            "absent_header",
            "1600000000.0",
            "{{config['min_wait']}}",
            None,
            SOME_BACKOFF_TIME,
        ),
    ],
)
@patch("time.time", return_value=1600000000.0)
def test_wait_untiltime_from_header(time_mock, test_name, header, wait_until, min_wait, regex, expected_backoff_time):
    response_mock = MagicMock()
    response_mock.headers = {"wait_until": wait_until}
    backoff_stratery = WaitUntilTimeFromHeaderBackoffStrategy(
        header=header,
        min_wait=min_wait,
        regex=regex,
        parameters={"wait_until": "wait_until", "regex": REGEX, "min_wait": SOME_BACKOFF_TIME},
        config={"wait_until": "wait_until", "regex": REGEX, "min_wait": SOME_BACKOFF_TIME},
    )
    backoff = backoff_stratery.backoff(response_mock, 1)
    assert backoff == expected_backoff_time



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import csv
import json
import logging
from functools import partial
from io import IOBase
from typing import Any, Callable, Dict, Iterable, List, Mapping, Optional, Set

from airbyte_cdk.sources.file_based.config.csv_format import CsvFormat, QuotingBehavior
from airbyte_cdk.sources.file_based.config.file_based_stream_config import FileBasedStreamConfig
from airbyte_cdk.sources.file_based.exceptions import FileBasedSourceError, RecordParseError
from airbyte_cdk.sources.file_based.file_based_stream_reader import AbstractFileBasedStreamReader, FileReadMode
from airbyte_cdk.sources.file_based.file_types.file_type_parser import FileTypeParser
from airbyte_cdk.sources.file_based.remote_file import RemoteFile
from airbyte_cdk.sources.file_based.schema_helpers import TYPE_PYTHON_MAPPING

DIALECT_NAME = "_config_dialect"

config_to_quoting: Mapping[QuotingBehavior, int] = {
    QuotingBehavior.QUOTE_ALL: csv.QUOTE_ALL,
    QuotingBehavior.QUOTE_SPECIAL_CHARACTERS: csv.QUOTE_MINIMAL,
    QuotingBehavior.QUOTE_NONNUMERIC: csv.QUOTE_NONNUMERIC,
    QuotingBehavior.QUOTE_NONE: csv.QUOTE_NONE,
}


class CsvParser(FileTypeParser):
    async def infer_schema(
        self,
        config: FileBasedStreamConfig,
        file: RemoteFile,
        stream_reader: AbstractFileBasedStreamReader,
        logger: logging.Logger,
    ) -> Dict[str, Any]:
        config_format = config.format.get(config.file_type) if config.format else CsvFormat()
        if not isinstance(config_format, CsvFormat):
            raise ValueError(f"Invalid format config: {config_format}")
        dialect_name = config.name + DIALECT_NAME
        csv.register_dialect(
            dialect_name,
            delimiter=config_format.delimiter,
            quotechar=config_format.quote_char,
            escapechar=config_format.escape_char,
            doublequote=config_format.double_quote,
            quoting=config_to_quoting.get(config_format.quoting_behavior, csv.QUOTE_MINIMAL),
        )
        with stream_reader.open_file(file, self.file_read_mode, logger) as fp:
            # todo: the existing InMemoryFilesSource.open_file() test source doesn't currently require an encoding, but actual
            #  sources will likely require one. Rather than modify the interface now we can wait until the real use case
            headers = self._get_headers(fp, config_format, dialect_name)
            schema = {field.strip(): {"type": "string"} for field in headers}
            csv.unregister_dialect(dialect_name)
            return schema

    def parse_records(
        self,
        config: FileBasedStreamConfig,
        file: RemoteFile,
        stream_reader: AbstractFileBasedStreamReader,
        logger: logging.Logger,
    ) -> Iterable[Dict[str, Any]]:
        schema: Mapping[str, Any] = config.input_schema  # type: ignore
        config_format = config.format.get(config.file_type) if config.format else CsvFormat()
        if not isinstance(config_format, CsvFormat):
            raise ValueError(f"Invalid format config: {config_format}")
        # Formats are configured individually per-stream so a unique dialect should be registered for each stream.
        # We don't unregister the dialect because we are lazily parsing each csv file to generate records
        # This will potentially be a problem if we ever process multiple streams concurrently
        dialect_name = config.name + DIALECT_NAME
        csv.register_dialect(
            dialect_name,
            delimiter=config_format.delimiter,
            quotechar=config_format.quote_char,
            escapechar=config_format.escape_char,
            doublequote=config_format.double_quote,
            quoting=config_to_quoting.get(config_format.quoting_behavior, csv.QUOTE_MINIMAL),
        )
        with stream_reader.open_file(file, self.file_read_mode, logger) as fp:
            # todo: the existing InMemoryFilesSource.open_file() test source doesn't currently require an encoding, but actual
            #  sources will likely require one. Rather than modify the interface now we can wait until the real use case
            self._skip_rows_before_header(fp, config_format.skip_rows_before_header)
            field_names = self._auto_generate_headers(fp, config_format) if config_format.autogenerate_column_names else None
            reader = csv.DictReader(fp, dialect=dialect_name, fieldnames=field_names)  # type: ignore
            yield from self._read_and_cast_types(reader, schema, config_format, logger)

    @property
    def file_read_mode(self) -> FileReadMode:
        return FileReadMode.READ

    @staticmethod
    def _read_and_cast_types(
        reader: csv.DictReader, schema: Optional[Mapping[str, Any]], config_format: CsvFormat, logger: logging.Logger  # type: ignore
    ) -> Iterable[Dict[str, Any]]:
        """
        If the user provided a schema, attempt to cast the record values to the associated type.

        If a column is not in the schema or cannot be cast to an appropriate python type,
        cast it to a string. Downstream, the user's validation policy will determine whether the
        record should be emitted.
        """
        cast_fn = CsvParser._get_cast_function(schema, config_format, logger)
        for i, row in enumerate(reader):
            if i < config_format.skip_rows_after_header:
                continue
            # The row was not properly parsed if any of  the values are None
            if any(val is None for val in row.values()):
                raise RecordParseError(FileBasedSourceError.ERROR_PARSING_RECORD)
            else:
                yield CsvParser._to_nullable(cast_fn(row), config_format.null_values)

    @staticmethod
    def _get_cast_function(
        schema: Optional[Mapping[str, Any]], config_format: CsvFormat, logger: logging.Logger
    ) -> Callable[[Mapping[str, str]], Mapping[str, str]]:
        # Only cast values if the schema is provided
        if schema:
            property_types = {col: prop["type"] for col, prop in schema["properties"].items()}
            return partial(_cast_types, property_types=property_types, config_format=config_format, logger=logger)
        else:
            # If no schema is provided, yield the rows as they are
            return _no_cast

    @staticmethod
    def _to_nullable(row: Mapping[str, str], null_values: Set[str]) -> Dict[str, Optional[str]]:
        nullable = row | {k: None if v in null_values else v for k, v in row.items()}
        return nullable

    @staticmethod
    def _skip_rows_before_header(fp: IOBase, rows_to_skip: int) -> None:
        """
        Skip rows before the header. This has to be done on the file object itself, not the reader
        """
        for _ in range(rows_to_skip):
            fp.readline()

    def _get_headers(self, fp: IOBase, config_format: CsvFormat, dialect_name: str) -> List[str]:
        # Note that this method assumes the dialect has already been registered if we're parsing the headers
        if config_format.autogenerate_column_names:
            return self._auto_generate_headers(fp, config_format)
        else:
            # If we're not autogenerating column names, we need to skip the rows before the header
            self._skip_rows_before_header(fp, config_format.skip_rows_before_header)
            # Then read the header
            reader = csv.DictReader(fp, dialect=dialect_name)  # type: ignore
            return next(reader)  # type: ignore

    def _auto_generate_headers(self, fp: IOBase, config_format: CsvFormat) -> List[str]:
        """
        Generates field names as [f0, f1, ...] in the same way as pyarrow's csv reader with autogenerate_column_names=True.
        See https://arrow.apache.org/docs/python/generated/pyarrow.csv.ReadOptions.html
        """
        next_line = next(fp).strip()
        number_of_columns = len(next_line.split(config_format.delimiter))  # type: ignore
        # Reset the file pointer to the beginning of the file so that the first row is not skipped
        fp.seek(0)
        return [f"f{i}" for i in range(number_of_columns)]


def _cast_types(row: Dict[str, str], property_types: Dict[str, Any], config_format: CsvFormat, logger: logging.Logger) -> Dict[str, Any]:
    """
    Casts the values in the input 'row' dictionary according to the types defined in the JSON schema.

    Array and object types are only handled if they can be deserialized as JSON.

    If any errors are encountered, the value will be emitted as a string.
    """
    warnings = []
    result = {}

    for key, value in row.items():
        prop_type = property_types.get(key)
        cast_value: Any = value

        if prop_type in TYPE_PYTHON_MAPPING and prop_type is not None:
            _, python_type = TYPE_PYTHON_MAPPING[prop_type]

            if python_type is None:
                if value == "":
                    cast_value = None
                else:
                    warnings.append(_format_warning(key, value, prop_type))

            elif python_type == bool:
                try:
                    cast_value = _value_to_bool(value, config_format.true_values, config_format.false_values)
                except ValueError:
                    warnings.append(_format_warning(key, value, prop_type))

            elif python_type == dict:
                try:
                    cast_value = json.loads(value)
                except json.JSONDecodeError:
                    warnings.append(_format_warning(key, value, prop_type))

            elif python_type == list:
                try:
                    parsed_value = json.loads(value)
                    if isinstance(parsed_value, list):
                        cast_value = parsed_value
                except json.JSONDecodeError:
                    warnings.append(_format_warning(key, value, prop_type))

            elif python_type:
                try:
                    cast_value = python_type(value)
                except ValueError:
                    warnings.append(_format_warning(key, value, prop_type))

        else:
            warnings.append(_format_warning(key, value, prop_type))

        result[key] = cast_value

    if warnings:
        logger.warning(
            f"{FileBasedSourceError.ERROR_CASTING_VALUE.value}: {','.join([w for w in warnings])}",
        )
    return result


def _value_to_bool(value: str, true_values: Set[str], false_values: Set[str]) -> bool:
    if value in true_values:
        return True
    if value in false_values:
        return False
    raise ValueError(f"Value {value} is not a valid boolean value")


def _format_warning(key: str, value: str, expected_type: Optional[Any]) -> str:
    return f"{key}: value={value},expected_type={expected_type}"


def _no_cast(row: Mapping[str, str]) -> Mapping[str, str]:
    return row



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#


import logging
from abc import ABC, abstractmethod
from typing import Generic, Iterable, Optional

from airbyte_cdk.connector import TConfig
from airbyte_cdk.models import AirbyteCatalog, AirbyteMessage, AirbyteStateMessage, ConfiguredAirbyteCatalog
from airbyte_cdk.sources.source import Source


class SourceRunner(ABC, Generic[TConfig]):
    @abstractmethod
    def discover(self, config: TConfig) -> AirbyteCatalog:
        pass

    @abstractmethod
    def read(self, config: TConfig, catalog: ConfiguredAirbyteCatalog, state: Optional[AirbyteStateMessage]) -> Iterable[AirbyteMessage]:
        pass


class CDKRunner(SourceRunner[TConfig]):
    def __init__(self, source: Source, name: str):
        self._source = source
        self._logger = logging.getLogger(name)

    def discover(self, config: TConfig) -> AirbyteCatalog:
        return self._source.discover(self._logger, config)

    def read(self, config: TConfig, catalog: ConfiguredAirbyteCatalog, state: Optional[AirbyteStateMessage]) -> Iterable[AirbyteMessage]:
        return self._source.read(self._logger, config, catalog, state=[state] if state else [])



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import pytest
from airbyte_cdk.utils.airbyte_secrets_utils import filter_secrets, get_secret_paths, get_secrets, update_secrets

SECRET_STRING_KEY = "secret_key1"
SECRET_STRING_VALUE = "secret_value"
SECRET_STRING_2_KEY = "secret_key2"
SECRET_STRING_2_VALUE = "second_secret_val"
SECRET_INT_KEY = "secret_int"
SECRET_INT_VALUE = 1337
NOT_SECRET_KEY = "not_a_secret"
NOT_SECRET_VALUE = "unimportant value"


flat_spec_with_secret = {"properties": {SECRET_STRING_KEY: {"type": "string", "airbyte_secret": True}, NOT_SECRET_KEY: {"type": "string"}}}
flat_config_with_secret = {SECRET_STRING_KEY: SECRET_STRING_VALUE, NOT_SECRET_KEY: NOT_SECRET_VALUE}

flat_spec_with_secret_int = {
    "properties": {SECRET_INT_KEY: {"type": "integer", "airbyte_secret": True}, NOT_SECRET_KEY: {"type": "string"}}
}
flat_config_with_secret_int = {SECRET_INT_KEY: SECRET_INT_VALUE, NOT_SECRET_KEY: NOT_SECRET_VALUE}

flat_spec_without_secrets = {"properties": {NOT_SECRET_KEY: {"type": "string"}}}
flat_config_without_secrets = {NOT_SECRET_KEY: NOT_SECRET_VALUE}

spec_with_oneof_secrets = {
    "properties": {
        SECRET_STRING_KEY: {"type": "string", "airbyte_secret": True},
        NOT_SECRET_KEY: {"type": "string"},
        "credentials": {
            "type": "object",
            "oneOf": [
                {
                    "type": "object",
                    "properties": {SECRET_STRING_2_KEY: {"type": "string", "airbyte_secret": True}, NOT_SECRET_KEY: {"type": "string"}},
                },
                {
                    "type": "object",
                    "properties": {SECRET_INT_KEY: {"type": "integer", "airbyte_secret": True}, NOT_SECRET_KEY: {"type": "string"}},
                },
            ],
        },
    }
}
config_with_oneof_secrets_1 = {
    SECRET_STRING_KEY: SECRET_STRING_VALUE,
    NOT_SECRET_KEY: NOT_SECRET_VALUE,
    "credentials": {SECRET_STRING_2_KEY: SECRET_STRING_2_VALUE},
}
config_with_oneof_secrets_2 = {
    SECRET_STRING_KEY: SECRET_STRING_VALUE,
    NOT_SECRET_KEY: NOT_SECRET_VALUE,
    "credentials": {SECRET_INT_KEY: SECRET_INT_VALUE},
}

spec_with_nested_secrets = {
    "properties": {
        SECRET_STRING_KEY: {"type": "string", "airbyte_secret": True},
        NOT_SECRET_KEY: {"type": "string"},
        "credentials": {
            "type": "object",
            "properties": {
                SECRET_STRING_2_KEY: {"type": "string", "airbyte_secret": True},
                NOT_SECRET_KEY: {"type": "string"},
                SECRET_INT_KEY: {"type": "integer", "airbyte_secret": True},
            },
        },
    }
}
config_with_nested_secrets = {
    SECRET_STRING_KEY: SECRET_STRING_VALUE,
    NOT_SECRET_KEY: NOT_SECRET_VALUE,
    "credentials": {SECRET_STRING_2_KEY: SECRET_STRING_2_VALUE, SECRET_INT_KEY: SECRET_INT_VALUE},
}


@pytest.mark.parametrize(
    ["spec", "expected"],
    [
        (flat_spec_with_secret, [[SECRET_STRING_KEY]]),
        (flat_spec_without_secrets, []),
        (flat_spec_with_secret_int, [[SECRET_INT_KEY]]),
        (spec_with_oneof_secrets, [[SECRET_STRING_KEY], ["credentials", SECRET_STRING_2_KEY], ["credentials", SECRET_INT_KEY]]),
        (spec_with_nested_secrets, [[SECRET_STRING_KEY], ["credentials", SECRET_STRING_2_KEY], ["credentials", SECRET_INT_KEY]]),
    ],
)
def test_get_secret_paths(spec, expected):
    assert get_secret_paths(spec) == expected, f"Expected {spec} to yield secret paths {expected}"


@pytest.mark.parametrize(
    ["spec", "config", "expected"],
    [
        (flat_spec_with_secret, flat_config_with_secret, [SECRET_STRING_VALUE]),
        (flat_spec_without_secrets, flat_config_without_secrets, []),
        (flat_spec_with_secret_int, flat_config_with_secret_int, [SECRET_INT_VALUE]),
        (spec_with_oneof_secrets, config_with_oneof_secrets_1, [SECRET_STRING_VALUE, SECRET_STRING_2_VALUE]),
        (spec_with_oneof_secrets, config_with_oneof_secrets_2, [SECRET_STRING_VALUE, SECRET_INT_VALUE]),
        (spec_with_nested_secrets, config_with_nested_secrets, [SECRET_STRING_VALUE, SECRET_STRING_2_VALUE, SECRET_INT_VALUE]),
    ],
)
def test_get_secrets(spec, config, expected):
    assert get_secrets(spec, config) == expected, f"Expected the spec {spec} and config {config} to produce {expected}"


def test_secret_filtering():
    sensitive_str = f"{SECRET_STRING_VALUE} {NOT_SECRET_VALUE} {SECRET_STRING_VALUE} {SECRET_STRING_2_VALUE}"

    update_secrets([])
    filtered = filter_secrets(sensitive_str)
    assert filtered == sensitive_str

    # the empty secret should not affect the result
    update_secrets([""])
    filtered = filter_secrets(sensitive_str)
    assert filtered == sensitive_str

    update_secrets([SECRET_STRING_VALUE, SECRET_STRING_2_VALUE])
    filtered = filter_secrets(sensitive_str)
    assert filtered == f"**** {NOT_SECRET_VALUE} **** ****"



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from abc import abstractmethod
from typing import Any, Mapping

from requests.auth import AuthBase


class AbstractHeaderAuthenticator(AuthBase):
    """Abstract class for an header-based authenticators that add a header to outgoing HTTP requests."""

    def __call__(self, request):
        """Attach the HTTP headers required to authenticate on the HTTP request"""
        request.headers.update(self.get_auth_header())
        return request

    def get_auth_header(self) -> Mapping[str, Any]:
        """The header to set on outgoing HTTP requests"""
        if self.auth_header:
            return {self.auth_header: self.token}
        return {}

    @property
    @abstractmethod
    def auth_header(self) -> str:
        """HTTP header to set on the requests"""

    @property
    @abstractmethod
    def token(self) -> str:
        """The header value to set on outgoing HTTP requests"""



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from datetime import datetime
from typing import Optional

from pydantic import BaseModel


class RemoteFile(BaseModel):
    """
    A file in a file-based stream.
    """

    uri: str
    last_modified: datetime

    def extension_agrees_with_file_type(self, file_type: Optional[str]) -> bool:
        extensions = self.uri.split(".")[1:]
        if not extensions:
            return True
        if not file_type:
            return True
        return any(file_type.casefold() in e.casefold() for e in extensions)



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from dataclasses import InitVar, dataclass
from typing import Any, Mapping


@dataclass
class RequestPath:
    """
    Describes that a component value should be inserted into the path
    """

    parameters: InitVar[Mapping[str, Any]]



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import pytest as pytest
from airbyte_cdk.sources.file_based.config.csv_format import CsvFormat


@pytest.mark.parametrize(
    "skip_rows_before_header, autogenerate_column_names, expected_error",
    [
        pytest.param(1, True, ValueError, id="test_skip_rows_before_header_and_autogenerate_column_names"),
        pytest.param(1, False, None, id="test_skip_rows_before_header_and_no_autogenerate_column_names"),
        pytest.param(0, True, None, id="test_no_skip_rows_before_header_and_autogenerate_column_names"),
        pytest.param(0, False, None, id="test_no_skip_rows_before_header_and_no_autogenerate_column_names"),
    ]
)
def test_csv_format(skip_rows_before_header, autogenerate_column_names, expected_error):
    if expected_error:
        with pytest.raises(expected_error):
            CsvFormat(skip_rows_before_header=skip_rows_before_header, autogenerate_column_names=autogenerate_column_names)
    else:
        CsvFormat(skip_rows_before_header=skip_rows_before_header, autogenerate_column_names=autogenerate_column_names)



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from airbyte_cdk.sources.file_based.discovery_policy.abstract_discovery_policy import AbstractDiscoveryPolicy

DEFAULT_N_CONCURRENT_REQUESTS = 10
DEFAULT_MAX_N_FILES_FOR_STREAM_SCHEMA_INFERENCE = 10


class DefaultDiscoveryPolicy(AbstractDiscoveryPolicy):
    """
    Default number of concurrent requests to send to the source on discover, and number
    of files to use for schema inference.
    """

    @property
    def n_concurrent_requests(self) -> int:
        return DEFAULT_N_CONCURRENT_REQUESTS

    @property
    def max_n_files_for_schema_inference(self) -> int:
        return DEFAULT_MAX_N_FILES_FOR_STREAM_SCHEMA_INFERENCE



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from abc import ABC, abstractmethod
from typing import Optional

from airbyte_cdk.sources.declarative.stream_slicers.stream_slicer import StreamSlicer
from airbyte_cdk.sources.declarative.types import Record, StreamSlice, StreamState


class Cursor(ABC, StreamSlicer):
    """
    Cursors are components that allow for incremental syncs. They keep track of what data has been consumed and slices the requests based on
    that information.
    """

    @abstractmethod
    def set_initial_state(self, stream_state: StreamState) -> None:
        """
        Cursors are not initialized with their state. As state is needed in order to function properly, this method should be called
        before calling anything else

        :param stream_state: The state of the stream as returned by get_stream_state
        """

    @abstractmethod
    def close_slice(self, stream_slice: StreamSlice, most_recent_record: Optional[Record]) -> None:
        """
        Update state based on the stream slice and the latest record. Note that `stream_slice.cursor_slice` and
        `last_record.associated_slice` are expected to be the same but we make it explicit here that `stream_slice` should be leveraged to
        update the state.

        :param stream_slice: slice to close
        :param last_record: the latest record we have received for the slice. This is important to consider because even if the cursor emits
          a slice, some APIs are not able to enforce the upper boundary. The outcome is that the last_record might have a higher cursor
          value than the slice upper boundary and if we want to reduce the duplication as much as possible, we need to consider the highest
          value between the internal cursor, the stream slice upper boundary and the record cursor value.
        """

    @abstractmethod
    def get_stream_state(self) -> StreamState:
        """
        Returns the current stream state. We would like to restrict it's usage since it does expose internal of state. As of 2023-06-14, it
        is used for two things:
        * Interpolation of the requests
        * Transformation of records
        * Saving the state

        For the first case, we are probably stuck with exposing the stream state. For the second, we can probably expose a method that
        allows for emitting the state to the platform.
        """

    @abstractmethod
    def should_be_synced(self, record: Record) -> bool:
        """
        Evaluating if a record should be synced allows for filtering and stop condition on pagination
        """

    @abstractmethod
    def is_greater_than_or_equal(self, first: Record, second: Record) -> bool:
        """
        Evaluating which record is greater in terms of cursor. This is used to avoid having to capture all the records to close a slice
        """



# The earlier versions of airbyte-cdk (0.28.0<=) had the airbyte_protocol python classes
# declared inline in the airbyte-cdk code. However, somewhere around Feb 2023 the
# Airbyte Protocol moved to its own repo/PyPi package, called airbyte-protocol-models.
# This directory including the airbyte_protocol.py and well_known_types.py files
# are just wrappers on top of that stand-alone package which do some namespacing magic
# to make the airbyte_protocol python classes available to the airbyte-cdk consumer as part
# of airbyte-cdk rather than a standalone package.
from .airbyte_protocol import (
    AdvancedAuth,
    AirbyteCatalog,
    AirbyteConnectionStatus,
    AirbyteControlConnectorConfigMessage,
    AirbyteControlMessage,
    AirbyteErrorTraceMessage,
    AirbyteEstimateTraceMessage,
    AirbyteGlobalState,
    AirbyteLogMessage,
    AirbyteMessage,
    AirbyteProtocol,
    AirbyteRecordMessage,
    AirbyteStateBlob,
    AirbyteStateMessage,
    AirbyteStateType,
    AirbyteStream,
    AirbyteStreamState,
    AirbyteStreamStatus,
    AirbyteStreamStatusTraceMessage,
    AirbyteTraceMessage,
    AuthFlowType,
    ConfiguredAirbyteCatalog,
    ConfiguredAirbyteStream,
    ConnectorSpecification,
    DestinationSyncMode,
    EstimateType,
    FailureType,
    Level,
    OAuthConfigSpecification,
    OrchestratorType,
    Status,
    StreamDescriptor,
    SyncMode,
    TraceType,
    Type,
)
from .well_known_types import (
    BinaryData,
    Boolean,
    Date,
    Integer,
    IntegerEnum,
    Model,
    Number,
    NumberEnum,
    String,
    TimestampWithoutTimezone,
    TimestampWithTimezone,
    TimeWithoutTimezone,
    TimeWithTimezone,
)



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from abc import ABC, abstractmethod
from typing import Generic, Iterable, Optional, TypeVar

from airbyte_cdk.connector import TConfig
from airbyte_cdk.sources.embedded.catalog import create_configured_catalog, get_stream, get_stream_names
from airbyte_cdk.sources.embedded.runner import SourceRunner
from airbyte_cdk.sources.embedded.tools import get_defined_id
from airbyte_protocol.models import AirbyteRecordMessage, AirbyteStateMessage, SyncMode, Type

TOutput = TypeVar("TOutput")


class BaseEmbeddedIntegration(ABC, Generic[TConfig, TOutput]):
    def __init__(self, runner: SourceRunner[TConfig], config: TConfig):
        self.source = runner
        self.config = config

        self.last_state: Optional[AirbyteStateMessage] = None

    @abstractmethod
    def _handle_record(self, record: AirbyteRecordMessage, id: Optional[str]) -> Optional[TOutput]:
        """
        Turn an Airbyte record into the appropriate output type for the integration.
        """
        pass

    def _load_data(self, stream_name: str, state: Optional[AirbyteStateMessage] = None) -> Iterable[TOutput]:
        catalog = self.source.discover(self.config)
        stream = get_stream(catalog, stream_name)
        if not stream:
            raise ValueError(f"Stream {stream_name} not found, the following streams are available: {', '.join(get_stream_names(catalog))}")
        if SyncMode.incremental not in stream.supported_sync_modes:
            configured_catalog = create_configured_catalog(stream, sync_mode=SyncMode.full_refresh)
        else:
            configured_catalog = create_configured_catalog(stream, sync_mode=SyncMode.incremental)

        for message in self.source.read(self.config, configured_catalog, state):
            if message.type == Type.RECORD:
                output = self._handle_record(message.record, get_defined_id(stream, message.record.data))
                if output:
                    yield output
            elif message.type is Type.STATE and message.state:
                self.last_state = message.state



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import logging
from unittest.mock import MagicMock, Mock

import pytest
from airbyte_cdk.sources.file_based.config.csv_format import DEFAULT_FALSE_VALUES, DEFAULT_TRUE_VALUES, CsvFormat
from airbyte_cdk.sources.file_based.exceptions import RecordParseError
from airbyte_cdk.sources.file_based.file_types.csv_parser import CsvParser, _cast_types

PROPERTY_TYPES = {
    "col1": "null",
    "col2": "boolean",
    "col3": "integer",
    "col4": "number",
    "col5": "string",
    "col6": "object",
    "col7": "array",
    "col8": "array",
    "col9": "array",
}

logger = logging.getLogger()


@pytest.mark.parametrize(
    "row, true_values, false_values, expected_output",
    [
        pytest.param(
            {
                "col1": "",
                "col2": "true",
                "col3": "1",
                "col4": "1.1",
                "col5": "asdf",
                "col6": '{"a": "b"}',
                "col7": '[1, 2]',
                "col8": '["1", "2"]',
                "col9": '[{"a": "b"}, {"a": "c"}]',
            },
            DEFAULT_TRUE_VALUES,
            DEFAULT_FALSE_VALUES,
            {
                "col1": None,
                "col2": True,
                "col3": 1,
                "col4": 1.1,
                "col5": "asdf",
                "col6": {"a": "b"},
                "col7": [1, 2],
                "col8": ["1", "2"],
                "col9": [{"a": "b"}, {"a": "c"}],
            }, id="cast-all-cols"),
        pytest.param({"col1": "1"}, DEFAULT_TRUE_VALUES, DEFAULT_FALSE_VALUES, {"col1": "1"}, id="cannot-cast-to-null"),
        pytest.param({"col2": "1"}, DEFAULT_TRUE_VALUES, DEFAULT_FALSE_VALUES, {"col2": True}, id="cast-1-to-bool"),
        pytest.param({"col2": "0"}, DEFAULT_TRUE_VALUES, DEFAULT_FALSE_VALUES, {"col2": False}, id="cast-0-to-bool"),
        pytest.param({"col2": "yes"}, DEFAULT_TRUE_VALUES, DEFAULT_FALSE_VALUES, {"col2": True}, id="cast-yes-to-bool"),
        pytest.param({"col2": "this_is_a_true_value"}, ["this_is_a_true_value"], DEFAULT_FALSE_VALUES, {"col2": True}, id="cast-custom-true-value-to-bool"),
        pytest.param({"col2": "this_is_a_false_value"}, DEFAULT_TRUE_VALUES, ["this_is_a_false_value"], {"col2": False}, id="cast-custom-false-value-to-bool"),
        pytest.param({"col2": "no"}, DEFAULT_TRUE_VALUES, DEFAULT_FALSE_VALUES, {"col2": False}, id="cast-no-to-bool"),
        pytest.param({"col2": "10"}, DEFAULT_TRUE_VALUES, DEFAULT_FALSE_VALUES, {"col2": "10"}, id="cannot-cast-to-bool"),
        pytest.param({"col3": "1.1"}, DEFAULT_TRUE_VALUES, DEFAULT_FALSE_VALUES, {"col3": "1.1"}, id="cannot-cast-to-int"),
        pytest.param({"col4": "asdf"}, DEFAULT_TRUE_VALUES, DEFAULT_FALSE_VALUES, {"col4": "asdf"}, id="cannot-cast-to-float"),
        pytest.param({"col6": "{'a': 'b'}"}, DEFAULT_TRUE_VALUES, DEFAULT_FALSE_VALUES, {"col6": "{'a': 'b'}"}, id="cannot-cast-to-dict"),
        pytest.param({"col7": "['a', 'b']"}, DEFAULT_TRUE_VALUES, DEFAULT_FALSE_VALUES, {"col7": "['a', 'b']"}, id="cannot-cast-to-list-of-ints"),
        pytest.param({"col8": "['a', 'b']"}, DEFAULT_TRUE_VALUES, DEFAULT_FALSE_VALUES, {"col8": "['a', 'b']"}, id="cannot-cast-to-list-of-strings"),
        pytest.param({"col9": "['a', 'b']"}, DEFAULT_TRUE_VALUES, DEFAULT_FALSE_VALUES, {"col9": "['a', 'b']"}, id="cannot-cast-to-list-of-objects"),
        pytest.param({"col10": "x"}, DEFAULT_TRUE_VALUES, DEFAULT_FALSE_VALUES, {"col10": "x"}, id="item-not-in-props-doesn't-error"),
    ]
)
def test_cast_to_python_type(row, true_values, false_values, expected_output):
    csv_format = CsvFormat(true_values=true_values, false_values=false_values)
    assert _cast_types(row, PROPERTY_TYPES, csv_format, logger) == expected_output


@pytest.mark.parametrize(
    "reader_values, expected_rows", [
        pytest.param([{"col1": "1", "col2": None}], None, id="raise_exception_if_any_value_is_none"),
        pytest.param([{"col1": "1", "col2": "2"}], [{"col1": "1", "col2": "2"}], id="read_no_cast"),
    ]
)
def test_read_and_cast_types(reader_values, expected_rows):
    reader = MagicMock()
    reader.__iter__.return_value = reader_values
    schema = {}
    config_format = CsvFormat()
    logger = Mock()

    parser = CsvParser()

    expected_rows = expected_rows
    if expected_rows is None:
        with pytest.raises(RecordParseError):
            list(parser._read_and_cast_types(reader, schema, config_format, logger))
    else:
        assert expected_rows == list(parser._read_and_cast_types(reader, schema, config_format, logger))



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import copy
from abc import abstractmethod
from typing import Any, Dict, List

from airbyte_cdk.sources.file_based.config.file_based_stream_config import FileBasedStreamConfig
from airbyte_cdk.sources.utils import schema_helpers
from pydantic import AnyUrl, BaseModel, Field


class AbstractFileBasedSpec(BaseModel):
    """
    Used during spec; allows the developer to configure the cloud provider specific options
    that are needed when users configure a file-based source.
    """

    streams: List[FileBasedStreamConfig] = Field(
        title="The list of streams to sync",
        description='Each instance of this configuration defines a <a href="https://docs.airbyte.com/cloud/core-concepts#stream">stream</a>. Use this to define which files belong in the stream, their format, and how they should be parsed and validated. When sending data to warehouse destination such as Snowflake or BigQuery, each stream is a separate table.',
        order=10,
    )

    @classmethod
    @abstractmethod
    def documentation_url(cls) -> AnyUrl:
        """
        :return: link to docs page for this source e.g. "https://docs.airbyte.com/integrations/sources/s3"
        """

    @classmethod
    def schema(cls, *args: Any, **kwargs: Any) -> Dict[str, Any]:
        """
        Generates the mapping comprised of the config fields
        """
        schema = super().schema(*args, **kwargs)
        transformed_schema = copy.deepcopy(schema)
        schema_helpers.expand_refs(transformed_schema)
        cls.replace_enum_allOf_and_anyOf(transformed_schema)
        cls.add_legacy_format(transformed_schema)

        return transformed_schema

    @staticmethod
    def replace_enum_allOf_and_anyOf(schema: Dict[str, Any]) -> Dict[str, Any]:
        """
        allOfs are not supported by the UI, but pydantic is automatically writing them for enums.
        Unpacks the enums under allOf and moves them up a level under the enum key
        anyOfs are also not supported by the UI, so we replace them with the similar oneOf, with the
        additional validation that an incoming config only matches exactly one of a field's types.
        """
        # this will need to add ["anyOf"] once we have more than one format type and loop over the list of elements
        objects_to_check = schema["properties"]["streams"]["items"]["properties"]["format"]
        if "additionalProperties" in objects_to_check:
            objects_to_check["additionalProperties"]["oneOf"] = objects_to_check["additionalProperties"].pop("anyOf", [])
            for format in objects_to_check["additionalProperties"]["oneOf"]:
                for key in format["properties"]:
                    object_property = format["properties"][key]
                    if "allOf" in object_property and "enum" in object_property["allOf"][0]:
                        object_property["enum"] = object_property["allOf"][0]["enum"]
                        object_property.pop("allOf")

        properties_to_change = ["primary_key", "input_schema"]
        for property_to_change in properties_to_change:
            schema["properties"]["streams"]["items"]["properties"][property_to_change]["oneOf"] = schema["properties"]["streams"]["items"][
                "properties"
            ][property_to_change].pop("anyOf")
        return schema

    @staticmethod
    def add_legacy_format(schema: Dict[str, Any]) -> Dict[str, Any]:
        """
        Because we still need to allow for configs using the legacy format (like source-s3) where file format options
        are at the top level and not mapped from file_type -> format options, the json schema used to validate the
        config must be adjusted to support the generic mapping object. Once configs no longer adhere to the old
        format we can remove this change.
        """
        legacy_format_options = {
            "title": "Legacy Format",
            # Explicitly require this field to make it mutually exclusive (oneOf) with the new format mapping file_type -> format
            "required": ["filetype"],
            "type": "object",
            "properties": {"filetype": {"title": "Filetype", "type": "string"}},
        }
        csv_format_options = schema["properties"]["streams"]["items"]["properties"]["format"]
        union_format = {"oneOf": [csv_format_options, legacy_format_options]}
        schema["properties"]["streams"]["items"]["properties"]["format"] = union_format
        return schema



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import json
import logging
from typing import Any, Dict, Iterable

from airbyte_cdk.sources.file_based.config.file_based_stream_config import FileBasedStreamConfig
from airbyte_cdk.sources.file_based.file_based_stream_reader import AbstractFileBasedStreamReader, FileReadMode
from airbyte_cdk.sources.file_based.file_types.file_type_parser import FileTypeParser
from airbyte_cdk.sources.file_based.remote_file import RemoteFile
from airbyte_cdk.sources.file_based.schema_helpers import PYTHON_TYPE_MAPPING, merge_schemas


class JsonlParser(FileTypeParser):

    MAX_BYTES_PER_FILE_FOR_SCHEMA_INFERENCE = 1_000_000

    async def infer_schema(
        self,
        config: FileBasedStreamConfig,
        file: RemoteFile,
        stream_reader: AbstractFileBasedStreamReader,
        logger: logging.Logger,
    ) -> Dict[str, Any]:
        """
        Infers the schema for the file by inferring the schema for each line, and merging
        it with the previously-inferred schema.
        """
        inferred_schema: Dict[str, Any] = {}
        read_bytes = 0

        with stream_reader.open_file(file, self.file_read_mode, logger) as fp:
            for line in fp:
                if read_bytes < self.MAX_BYTES_PER_FILE_FOR_SCHEMA_INFERENCE:
                    line_schema = self.infer_schema_for_record(json.loads(line))
                    inferred_schema = merge_schemas(inferred_schema, line_schema)
                    read_bytes += len(line)

        if read_bytes > self.MAX_BYTES_PER_FILE_FOR_SCHEMA_INFERENCE:
            logger.warning(
                f"Exceeded the maximum number of bytes per file for schema inference ({self.MAX_BYTES_PER_FILE_FOR_SCHEMA_INFERENCE}). "
                f"Inferring schema from an incomplete set of records."
            )

        return inferred_schema

    def parse_records(
        self,
        config: FileBasedStreamConfig,
        file: RemoteFile,
        stream_reader: AbstractFileBasedStreamReader,
        logger: logging.Logger,
    ) -> Iterable[Dict[str, Any]]:
        with stream_reader.open_file(file, self.file_read_mode, logger) as fp:
            for line in fp:
                yield json.loads(line)

    @classmethod
    def infer_schema_for_record(cls, record: Dict[str, Any]) -> Dict[str, Any]:
        record_schema = {}
        for key, value in record.items():
            if value is None:
                record_schema[key] = {"type": "null"}
            else:
                record_schema[key] = {"type": PYTHON_TYPE_MAPPING[type(value)]}

        return record_schema

    @property
    def file_read_mode(self) -> FileReadMode:
        return FileReadMode.READ



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import logging
from typing import Any, Iterable, Mapping, Optional
from unittest.mock import MagicMock

import pytest
import requests
from airbyte_cdk.sources.declarative.checks.check_stream import CheckStream
from airbyte_cdk.sources.streams.http import HttpStream
from airbyte_cdk.sources.streams.http.availability_strategy import HttpAvailabilityStrategy

logger = logging.getLogger("test")
config = dict()

stream_names = ["s1"]
record = MagicMock()


@pytest.mark.parametrize(
    "test_name, record, streams_to_check, stream_slice, expectation",
    [
        ("test_success_check", record, stream_names, {}, (True, None)),
        ("test_success_check_stream_slice", record, stream_names, {"slice": "slice_value"}, (True, None)),
        ("test_fail_check", None, stream_names, {}, (True, None)),
        ("test_try_to_check_invalid stream", record, ["invalid_stream_name"], {}, None),
    ],
)
@pytest.mark.parametrize("slices_as_list", [True, False])
def test_check_stream_with_slices_as_list(test_name, record, streams_to_check, stream_slice, expectation, slices_as_list):
    stream = MagicMock()
    stream.name = "s1"
    stream.availability_strategy = None
    if slices_as_list:
        stream.stream_slices.return_value = [stream_slice]
    else:
        stream.stream_slices.return_value = iter([stream_slice])

    stream.read_records.side_effect = mock_read_records({frozenset(stream_slice): iter([record])})

    source = MagicMock()
    source.streams.return_value = [stream]

    check_stream = CheckStream(streams_to_check, parameters={})

    if expectation:
        actual = check_stream.check_connection(source, logger, config)
        assert actual == expectation
    else:
        with pytest.raises(ValueError):
            check_stream.check_connection(source, logger, config)


def mock_read_records(responses, default_response=None, **kwargs):
    return lambda stream_slice, sync_mode: responses[frozenset(stream_slice)] if frozenset(stream_slice) in responses else default_response


def test_check_empty_stream():
    stream = MagicMock()
    stream.name = "s1"
    stream.availability_strategy = None
    stream.read_records.return_value = iter([])
    stream.stream_slices.return_value = iter([None])

    source = MagicMock()
    source.streams.return_value = [stream]

    check_stream = CheckStream(["s1"], parameters={})
    stream_is_available, reason = check_stream.check_connection(source, logger, config)
    assert stream_is_available


def test_check_stream_with_no_stream_slices_aborts():
    stream = MagicMock()
    stream.name = "s1"
    stream.availability_strategy = None
    stream.stream_slices.return_value = iter([])

    source = MagicMock()
    source.streams.return_value = [stream]

    check_stream = CheckStream(["s1"], parameters={})
    stream_is_available, reason = check_stream.check_connection(source, logger, config)
    assert not stream_is_available
    assert "no stream slices were found, likely because the parent stream is empty" in reason


@pytest.mark.parametrize(
    "test_name, response_code, available_expectation, expected_messages",
    [
        ("test_stream_unavailable_unhandled_error", 404, False, ["Unable to connect to stream mock_http_stream", "404 Client Error"]),
        (
            "test_stream_unavailable_handled_error",
            403,
            False,
            [
                "Unable to read mock_http_stream stream",
                "This is most likely due to insufficient permissions on the credentials in use.",
            ],
        ),
        ("test_stream_available", 200, True, []),
    ],
)
def test_check_http_stream_via_availability_strategy(mocker, test_name, response_code, available_expectation, expected_messages):
    class MockHttpStream(HttpStream):
        url_base = "https://test_base_url.com"
        primary_key = ""

        def __init__(self, **kwargs):
            super().__init__(**kwargs)
            self.resp_counter = 1

        def next_page_token(self, response: requests.Response) -> Optional[Mapping[str, Any]]:
            return None

        def path(self, **kwargs) -> str:
            return ""

        def parse_response(self, response: requests.Response, **kwargs) -> Iterable[Mapping]:
            stub_resp = {"data": self.resp_counter}
            self.resp_counter += 1
            yield stub_resp

        pass

    http_stream = MockHttpStream()
    assert isinstance(http_stream, HttpStream)
    assert isinstance(http_stream.availability_strategy, HttpAvailabilityStrategy)

    source = MagicMock()
    source.streams.return_value = [http_stream]

    check_stream = CheckStream(stream_names=["mock_http_stream"], parameters={})

    req = requests.Response()
    req.status_code = response_code
    mocker.patch.object(requests.Session, "send", return_value=req)

    logger = logging.getLogger(f"airbyte.{getattr(source, 'name', '')}")
    stream_is_available, reason = check_stream.check_connection(source, logger, config)

    assert stream_is_available == available_expectation
    for message in expected_messages:
        assert message in reason



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from dataclasses import InitVar, dataclass
from typing import Any, List, Mapping, Optional

from airbyte_cdk.sources.declarative.interpolation.interpolated_boolean import InterpolatedBoolean
from airbyte_cdk.sources.declarative.types import Config, StreamSlice, StreamState


@dataclass
class RecordFilter:
    """
    Filter applied on a list of Records

    config (Config): The user-provided configuration as specified by the source's spec
    condition (str): The string representing the predicate to filter a record. Records will be removed if evaluated to False
    """

    parameters: InitVar[Mapping[str, Any]]
    config: Config
    condition: str = ""

    def __post_init__(self, parameters: Mapping[str, Any]) -> None:
        self._filter_interpolator = InterpolatedBoolean(condition=self.condition, parameters=parameters)

    def filter_records(
        self,
        records: List[Mapping[str, Any]],
        stream_state: StreamState,
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> List[Mapping[str, Any]]:
        kwargs = {"stream_state": stream_state, "stream_slice": stream_slice, "next_page_token": next_page_token}
        return [record for record in records if self._filter_interpolator.eval(self.config, record=record, **kwargs)]



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from collections import OrderedDict
from unittest.mock import Mock

import pytest
from airbyte_cdk.sources.declarative.incremental.cursor import Cursor
from airbyte_cdk.sources.declarative.incremental.per_partition_cursor import (
    PerPartitionCursor,
    PerPartitionKeySerializer,
    PerPartitionStreamSlice,
)
from airbyte_cdk.sources.declarative.stream_slicers.stream_slicer import StreamSlicer
from airbyte_cdk.sources.declarative.types import Record

PARTITION = {
    "partition_key string": "partition value",
    "partition_key int": 1,
    "partition_key list str": ["list item 1", "list item 2"],
    "partition_key list dict": [
        {
            "dict within list key 1-1": "dict within list value 1-1",
            "dict within list key 1-2": "dict within list value 1-2"
        },
        {"dict within list key 2": "dict within list value 2"},
    ],
    "partition_key nested dict": {
        "nested_partition_key 1": "a nested value",
        "nested_partition_key 2": "another nested value",
    },
}

CURSOR_SLICE_FIELD = "cursor slice field"
CURSOR_STATE_KEY = "cursor state"
CURSOR_STATE = {CURSOR_STATE_KEY: "a state value"}
NOT_CONSIDERED_BECAUSE_MOCKED_CURSOR_HAS_NO_STATE = "any"
STATE = {
    "states": [
        {
            "partition": {
                "partition_router_field_1": "X1",
                "partition_router_field_2": "Y1",
            },
            "cursor": {
                "cursor state field": 1
            }
        },
        {
            "partition": {
                "partition_router_field_1": "X2",
                "partition_router_field_2": "Y2",
            },
            "cursor": {
                "cursor state field": 2
            }
        },
    ]
}


def test_partition_serialization():
    serializer = PerPartitionKeySerializer()
    assert serializer.to_partition(serializer.to_partition_key(PARTITION)) == PARTITION


def test_partition_with_different_key_orders():
    ordered_dict = OrderedDict({"1": 1, "2": 2})
    same_dict_with_different_order = OrderedDict({"2": 2, "1": 1})
    serializer = PerPartitionKeySerializer()

    assert serializer.to_partition_key(ordered_dict) == serializer.to_partition_key(same_dict_with_different_order)


def test_given_tuples_in_json_then_deserialization_convert_to_list():
    """
    This is a known issue with the current implementation. However, the assumption is that this wouldn't be a problem as we only use the
    immutability and we expect stream slices to be immutable anyway
    """
    serializer = PerPartitionKeySerializer()
    partition_with_tuple = {"key": (1, 2, 3)}

    assert partition_with_tuple != serializer.to_partition(serializer.to_partition_key(partition_with_tuple))


def test_stream_slice_merge_dictionaries():
    stream_slice = PerPartitionStreamSlice({"partition key": "partition value"}, {"cursor key": "cursor value"})
    assert stream_slice == {"partition key": "partition value", "cursor key": "cursor value"}


def test_overlapping_slice_keys_raise_error():
    with pytest.raises(ValueError):
        PerPartitionStreamSlice({"overlapping key": "partition value"}, {"overlapping key": "cursor value"})


class MockedCursorBuilder:
    def __init__(self):
        self._stream_slices = []
        self._stream_state = {}

    def with_stream_slices(self, stream_slices):
        self._stream_slices = stream_slices
        return self

    def with_stream_state(self, stream_state):
        self._stream_state = stream_state
        return self

    def build(self):
        cursor = Mock(spec=Cursor)
        cursor.get_stream_state.return_value = self._stream_state
        cursor.stream_slices.return_value = self._stream_slices
        return cursor


@pytest.fixture()
def mocked_partition_router():
    return Mock(spec=StreamSlicer)


@pytest.fixture()
def mocked_cursor_factory():
    cursor_factory = Mock()
    cursor_factory.create.return_value = MockedCursorBuilder().build()
    return cursor_factory


def test_given_no_partition_when_stream_slices_then_no_slices(mocked_cursor_factory, mocked_partition_router):
    mocked_partition_router.stream_slices.return_value = []
    cursor = PerPartitionCursor(mocked_cursor_factory, mocked_partition_router)

    slices = cursor.stream_slices()

    assert not next(slices, None)


def test_given_partition_router_without_state_has_one_partition_then_return_one_slice_per_cursor_slice(mocked_cursor_factory,
                                                                                                       mocked_partition_router):
    partition = {"partition_field_1": "a value", "partition_field_2": "another value"}
    mocked_partition_router.stream_slices.return_value = [partition]
    cursor_slices = [{"start_datetime": 1}, {"start_datetime": 2}]
    mocked_cursor_factory.create.return_value = MockedCursorBuilder().with_stream_slices(cursor_slices).build()
    cursor = PerPartitionCursor(mocked_cursor_factory, mocked_partition_router)

    slices = cursor.stream_slices()

    assert list(slices) == [PerPartitionStreamSlice(partition, cursor_slice) for cursor_slice in cursor_slices]


def test_given_partition_associated_with_state_when_stream_slices_then_do_not_recreate_cursor(mocked_cursor_factory,
                                                                                              mocked_partition_router):
    partition = {"partition_field_1": "a value", "partition_field_2": "another value"}
    mocked_partition_router.stream_slices.return_value = [partition]
    cursor_slices = [{"start_datetime": 1}]
    mocked_cursor_factory.create.return_value = MockedCursorBuilder().with_stream_slices(cursor_slices).build()
    cursor = PerPartitionCursor(mocked_cursor_factory, mocked_partition_router)

    cursor.set_initial_state({
        "states": [{
            "partition": partition,
            "cursor": CURSOR_STATE
        }]
    })
    mocked_cursor_factory.create.assert_called_once()
    slices = list(cursor.stream_slices())

    mocked_cursor_factory.create.assert_called_once()
    assert len(slices) == 1


def test_given_multiple_partitions_then_each_have_their_state(mocked_cursor_factory, mocked_partition_router):
    first_partition = {"first_partition_key": "first_partition_value"}
    mocked_partition_router.stream_slices.return_value = [
        first_partition,
        {"second_partition_key": "second_partition_value"}
    ]
    first_cursor = MockedCursorBuilder().with_stream_slices([{CURSOR_SLICE_FIELD: "first slice cursor value"}]).build()
    second_cursor = MockedCursorBuilder().with_stream_slices([{CURSOR_SLICE_FIELD: "second slice cursor value"}]).build()
    mocked_cursor_factory.create.side_effect = [first_cursor, second_cursor]
    cursor = PerPartitionCursor(mocked_cursor_factory, mocked_partition_router)

    cursor.set_initial_state({
        "states": [{
            "partition": first_partition,
            "cursor": CURSOR_STATE
        }]
    })
    slices = list(cursor.stream_slices())

    first_cursor.stream_slices.assert_called_once()
    second_cursor.stream_slices.assert_called_once()
    assert slices == [
        PerPartitionStreamSlice(
            partition={"first_partition_key": "first_partition_value"},
            cursor_slice={CURSOR_SLICE_FIELD: "first slice cursor value"}
        ),
        PerPartitionStreamSlice(
            partition={"second_partition_key": "second_partition_value"},
            cursor_slice={CURSOR_SLICE_FIELD: "second slice cursor value"}
        ),
    ]


def test_given_stream_slices_when_get_stream_state_then_return_updated_state(mocked_cursor_factory, mocked_partition_router):
    mocked_cursor_factory.create.side_effect = [
        MockedCursorBuilder().with_stream_state({CURSOR_STATE_KEY: "first slice cursor value"}).build(),
        MockedCursorBuilder().with_stream_state({CURSOR_STATE_KEY: "second slice cursor value"}).build()
    ]
    mocked_partition_router.stream_slices.return_value = [{"partition key": "first partition"}, {"partition key": "second partition"}]
    cursor = PerPartitionCursor(mocked_cursor_factory, mocked_partition_router)
    list(cursor.stream_slices())
    assert cursor.get_stream_state() == {
        "states": [
            {
                "partition": {"partition key": "first partition"},
                "cursor": {CURSOR_STATE_KEY: "first slice cursor value"}
            },
            {
                "partition": {"partition key": "second partition"},
                "cursor": {CURSOR_STATE_KEY: "second slice cursor value"}
            }
        ]
    }


def test_when_get_stream_state_then_delegate_to_underlying_cursor(mocked_cursor_factory, mocked_partition_router):
    underlying_cursor = MockedCursorBuilder().with_stream_slices([{CURSOR_SLICE_FIELD: "first slice cursor value"}]).build()
    mocked_cursor_factory.create.side_effect = [underlying_cursor]
    mocked_partition_router.stream_slices.return_value = [{"partition key": "first partition"}]
    cursor = PerPartitionCursor(mocked_cursor_factory, mocked_partition_router)
    first_slice = list(cursor.stream_slices())[0]

    cursor.should_be_synced(
        Record(
            {},
            first_slice
        )
    )

    underlying_cursor.should_be_synced.assert_called_once_with(
        Record(
            {},
            first_slice.cursor_slice
        )
    )


def test_close_slice(mocked_cursor_factory, mocked_partition_router):
    underlying_cursor = MockedCursorBuilder().with_stream_slices([{CURSOR_SLICE_FIELD: "first slice cursor value"}]).build()
    mocked_cursor_factory.create.side_effect = [underlying_cursor]
    stream_slice = PerPartitionStreamSlice(partition={"partition key": "first partition"}, cursor_slice={})
    mocked_partition_router.stream_slices.return_value = [stream_slice.partition]
    cursor = PerPartitionCursor(mocked_cursor_factory, mocked_partition_router)
    last_record = Mock()
    list(cursor.stream_slices())  # generate internal state

    cursor.close_slice(stream_slice, last_record)

    underlying_cursor.close_slice.assert_called_once_with(stream_slice.cursor_slice, Record(last_record.data, stream_slice.cursor_slice))


def test_given_no_last_record_when_close_slice_then_do_not_raise_error(mocked_cursor_factory, mocked_partition_router):
    underlying_cursor = MockedCursorBuilder().with_stream_slices([{CURSOR_SLICE_FIELD: "first slice cursor value"}]).build()
    mocked_cursor_factory.create.side_effect = [underlying_cursor]
    stream_slice = PerPartitionStreamSlice(partition={"partition key": "first partition"}, cursor_slice={})
    mocked_partition_router.stream_slices.return_value = [stream_slice.partition]
    cursor = PerPartitionCursor(mocked_cursor_factory, mocked_partition_router)
    list(cursor.stream_slices())  # generate internal state

    cursor.close_slice(stream_slice, None)

    underlying_cursor.close_slice.assert_called_once_with(stream_slice.cursor_slice, None)


def test_given_unknown_partition_when_close_slice_then_raise_error():
    any_cursor_factory = Mock()
    any_partition_router = Mock()
    cursor = PerPartitionCursor(any_cursor_factory, any_partition_router)
    stream_slice = PerPartitionStreamSlice(partition={"unknown_partition": "unknown"}, cursor_slice={})
    with pytest.raises(ValueError):
        cursor.close_slice(
            stream_slice,
            Record({}, stream_slice)
        )


def test_given_unknown_partition_when_should_be_synced_then_raise_error():
    any_cursor_factory = Mock()
    any_partition_router = Mock()
    cursor = PerPartitionCursor(any_cursor_factory, any_partition_router)
    with pytest.raises(ValueError):
        cursor.should_be_synced(
            Record(
                {},
                PerPartitionStreamSlice(
                    partition={"unknown_partition": "unknown"},
                    cursor_slice={}
                )
            )
        )


def test_given_records_with_different_slice_when_is_greater_than_or_equal_then_raise_error():
    any_cursor_factory = Mock()
    any_partition_router = Mock()
    cursor = PerPartitionCursor(any_cursor_factory, any_partition_router)
    with pytest.raises(ValueError):
        cursor.is_greater_than_or_equal(
            Record({}, PerPartitionStreamSlice(partition={"a slice": "value"}, cursor_slice={})),
            Record({}, PerPartitionStreamSlice(partition={"another slice": "value"}, cursor_slice={}))
        )


def test_given_slice_is_unknown_when_is_greater_than_or_equal_then_raise_error():
    any_cursor_factory = Mock()
    any_partition_router = Mock()
    cursor = PerPartitionCursor(any_cursor_factory, any_partition_router)
    with pytest.raises(ValueError):
        cursor.is_greater_than_or_equal(
            Record({}, PerPartitionStreamSlice(partition={"a slice": "value"}, cursor_slice={})),
            Record({}, PerPartitionStreamSlice(partition={"a slice": "value"}, cursor_slice={}))
        )


def test_when_is_greater_than_or_equal_then_return_underlying_cursor_response(mocked_cursor_factory, mocked_partition_router):
    underlying_cursor = MockedCursorBuilder().with_stream_slices([{CURSOR_SLICE_FIELD: "first slice cursor value"}]).build()
    mocked_cursor_factory.create.side_effect = [underlying_cursor]
    stream_slice = PerPartitionStreamSlice(partition={"partition key": "first partition"}, cursor_slice={})
    mocked_partition_router.stream_slices.return_value = [stream_slice.partition]
    cursor = PerPartitionCursor(mocked_cursor_factory, mocked_partition_router)
    first_record = Record({"first": "value"}, stream_slice)
    second_record = Record({"second": "value"}, stream_slice)
    list(cursor.stream_slices())  # generate internal state

    result = cursor.is_greater_than_or_equal(first_record, second_record)

    assert result == underlying_cursor.is_greater_than_or_equal.return_value
    underlying_cursor.is_greater_than_or_equal.assert_called_once_with(first_record, second_record)



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import datetime
import decimal
import uuid

from unit_tests.sources.file_based.in_memory_files_source import TemporaryAvroFilesStreamReader
from unit_tests.sources.file_based.scenarios.scenario_builder import TestScenarioBuilder

_single_avro_file = {
    "a.avro": {
        "schema": {
            "type": "record",
            "name": "sampleAvro",
            "fields": [
                {"name": "col1", "type": "string"},
                {"name": "col2", "type": "int"},
            ],
        },
        "contents": [
            ("val11", 12),
            ("val21", 22),
        ],
        "last_modified": "2023-06-05T03:54:07.000Z",
    }
}

_multiple_avro_combine_schema_file = {
    "a.avro": {
        "schema": {
            "type": "record",
            "name": "sampleAvro",
            "fields": [
                {"name": "col_double", "type": "double"},
                {"name": "col_string", "type": "string"},
                {"name": "col_album", "type": {"type": "record", "name": "Album", "fields": [{"name": "album", "type": "string"}]}},
            ],
        },
        "contents": [
            (20.02, "Robbers", {"album": "The 1975"}),
            (20.23, "Somebody Else", {"album": "I Like It When You Sleep, for You Are So Beautiful yet So Unaware of It"}),
        ],
        "last_modified": "2023-06-05T03:54:07.000Z",
    },
    "b.avro": {
        "schema": {
            "type": "record",
            "name": "sampleAvro",
            "fields": [
                {"name": "col_double", "type": "double"},
                {"name": "col_string", "type": "string"},
                {"name": "col_song", "type": {"type": "record", "name": "Song", "fields": [{"name": "title", "type": "string"}]}},
            ],
        },
        "contents": [
            (1975.1975, "It's Not Living (If It's Not with You)", {"title": "Love It If We Made It"}),
            (5791.5791, "The 1975", {"title": "About You"}),
        ],
        "last_modified": "2023-06-06T03:54:07.000Z",
    },
}

_avro_all_types_file = {
    "a.avro": {
        "schema": {
            "type": "record",
            "name": "sampleAvro",
            "fields": [
                # Primitive Types
                {"name": "col_bool", "type": "boolean"},
                {"name": "col_int", "type": "int"},
                {"name": "col_long", "type": "long"},
                {"name": "col_float", "type": "float"},
                {"name": "col_double", "type": "double"},
                {"name": "col_bytes", "type": "bytes"},
                {"name": "col_string", "type": "string"},
                # Complex Types
                {
                    "name": "col_record",
                    "type": {
                        "type": "record",
                        "name": "SongRecord",
                        "fields": [
                            {"name": "artist", "type": "string"},
                            {"name": "song", "type": "string"},
                            {"name": "year", "type": "int"},
                        ],
                    },
                },
                {"name": "col_enum", "type": {"type": "enum", "name": "Genre", "symbols": ["POP_ROCK", "INDIE_ROCK", "ALTERNATIVE_ROCK"]}},
                {"name": "col_array", "type": {"type": "array", "items": "string"}},
                {"name": "col_map", "type": {"type": "map", "values": "string"}},
                {"name": "col_fixed", "type": {"type": "fixed", "name": "MyFixed", "size": 4}},
                # Logical Types
                {"name": "col_decimal", "type": {"type": "bytes", "logicalType": "decimal", "precision": 10, "scale": 5}},
                {"name": "col_uuid", "type": {"type": "bytes", "logicalType": "uuid"}},
                {"name": "col_date", "type": {"type": "int", "logicalType": "date"}},
                {"name": "col_time_millis", "type": {"type": "int", "logicalType": "time-millis"}},
                {"name": "col_time_micros", "type": {"type": "long", "logicalType": "time-micros"}},
                {"name": "col_timestamp_millis", "type": {"type": "long", "logicalType": "timestamp-millis"}},
                {"name": "col_timestamp_micros", "type": {"type": "long", "logicalType": "timestamp-micros"}},
            ],
        },
        "contents": [
            (
                True,
                27,
                1992,
                999.09723456,
                9123456.12394,
                b"\x00\x01\x02\x03",
                "Love It If We Made It",
                {"artist": "The 1975", "song": "About You", "year": 2022},
                "POP_ROCK",
                [
                    "The 1975",
                    "I Like It When You Sleep, for You Are So Beautiful yet So Unaware of It",
                    "The 1975 A Brief Inquiry into Online Relationships",
                    "Notes on a Conditional Form",
                    "Being Funny in a Foreign Language",
                ],
                {"lead_singer": "Matty Healy", "lead_guitar": "Adam Hann", "bass_guitar": "Ross MacDonald", "drummer": "George Daniel"},
                b"\x12\x34\x56\x78",
                decimal.Decimal("1234.56789"),
                uuid.UUID('123e4567-e89b-12d3-a456-426655440000').bytes,
                datetime.date(2022, 5, 29),
                datetime.time(6, 0, 0, 456000),
                datetime.time(12, 0, 0, 456789),
                datetime.datetime(2022, 5, 29, 0, 0, 0, 456000, tzinfo=datetime.timezone.utc),
                datetime.datetime(2022, 5, 30, 0, 0, 0, 456789, tzinfo=datetime.timezone.utc),
            ),
        ],
        "last_modified": "2023-06-05T03:54:07.000Z",
    }
}

_multiple_avro_stream_file = {
    "odesza_songs.avro": {
        "schema": {
            "type": "record",
            "name": "sampleAvro",
            "fields": [
                {"name": "col_title", "type": "string"},
                {
                    "name": "col_album",
                    "type": {
                        "type": "enum",
                        "name": "Album",
                        "symbols": ["SUMMERS_GONE", "IN_RETURN", "A_MOMENT_APART", "THE_LAST_GOODBYE"],
                    },
                },
                {"name": "col_year", "type": "int"},
                {"name": "col_vocals", "type": "boolean"},
            ],
        },
        "contents": [
            ("Late Night", "A_MOMENT_APART", 2017, False),
            ("White Lies", "IN_RETURN", 2014, True),
            ("Wide Awake", "THE_LAST_GOODBYE", 2022, True),
            ("Sun Models", "SUMMERS_GONE", 2012, True),
            ("All We Need", "IN_RETURN", 2014, True),
        ],
        "last_modified": "2023-06-05T03:54:07.000Z",
    },
    "california_festivals.avro": {
        "schema": {
            "type": "record",
            "name": "sampleAvro",
            "fields": [
                {"name": "col_name", "type": "string"},
                {
                    "name": "col_location",
                    "type": {
                        "type": "record",
                        "name": "LocationRecord",
                        "fields": [
                            {"name": "country", "type": "string"},
                            {"name": "state", "type": "string"},
                            {"name": "city", "type": "string"},
                        ],
                    },
                },
                {"name": "col_attendance", "type": "long"},
            ],
        },
        "contents": [
            ("Coachella", {"country": "USA", "state": "California", "city": "Indio"}, 250000),
            ("CRSSD", {"country": "USA", "state": "California", "city": "San Diego"}, 30000),
            ("Lightning in a Bottle", {"country": "USA", "state": "California", "city": "Buena Vista Lake"}, 18000),
            ("Outside Lands", {"country": "USA", "state": "California", "city": "San Francisco"}, 220000),
        ],
        "last_modified": "2023-06-06T03:54:07.000Z",
    },
}

single_avro_scenario = (
    TestScenarioBuilder()
    .set_name("single_avro_stream")
    .set_config(
        {
            "streams": [
                {
                    "name": "stream1",
                    "file_type": "avro",
                    "globs": ["*"],
                    "validation_policy": "emit_record",
                }
            ]
        }
    )
    .set_stream_reader(TemporaryAvroFilesStreamReader(files=_single_avro_file, file_type="avro"))
    .set_file_type("avro")
    .set_expected_check_status("SUCCEEDED")
    .set_expected_records(
        [
            {
                "data": {
                    "col1": "val11",
                    "col2": 12,
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a.avro",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val21",
                    "col2": 22,
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a.avro",
                },
                "stream": "stream1",
            },
        ]
    )
    .set_expected_catalog(
        {
            "streams": [
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col1": {"type": ["null", "string"]},
                            "col2": {"type": ["null", "integer"]},
                            "_ab_source_file_last_modified": {"type": "string"},
                            "_ab_source_file_url": {"type": "string"},
                        },
                    },
                    "name": "stream1",
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                }
            ]
        }
    )
).build()

multiple_avro_combine_schema_scenario = (
    TestScenarioBuilder()
    .set_name("multiple_avro_combine_schema_stream")
    .set_config(
        {
            "streams": [
                {
                    "name": "stream1",
                    "file_type": "avro",
                    "globs": ["*"],
                    "validation_policy": "emit_record",
                }
            ]
        }
    )
    .set_stream_reader(TemporaryAvroFilesStreamReader(files=_multiple_avro_combine_schema_file, file_type="avro"))
    .set_file_type("avro")
    .set_expected_records(
        [
            {
                "data": {
                    "col_double": "20.02",
                    "col_string": "Robbers",
                    "col_album": {"album": "The 1975"},
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a.avro",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col_double": "20.23",
                    "col_string": "Somebody Else",
                    "col_album": {"album": "I Like It When You Sleep, for You Are So Beautiful yet So Unaware of It"},
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a.avro",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col_double": "1975.1975",
                    "col_string": "It's Not Living (If It's Not with You)",
                    "col_song": {"title": "Love It If We Made It"},
                    "_ab_source_file_last_modified": "2023-06-06T03:54:07.000000Z",
                    "_ab_source_file_url": "b.avro",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col_double": "5791.5791",
                    "col_string": "The 1975",
                    "col_song": {"title": "About You"},
                    "_ab_source_file_last_modified": "2023-06-06T03:54:07.000000Z",
                    "_ab_source_file_url": "b.avro",
                },
                "stream": "stream1",
            },
        ]
    )
    .set_expected_catalog(
        {
            "streams": [
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col_double": {"type": ["null", "string"]},
                            "col_string": {"type": ["null", "string"]},
                            "col_album": {
                                "properties": {
                                    "album": {"type": ["null", "string"]},
                                },
                                "type": ["null", "object"],
                            },
                            "col_song": {
                                "properties": {
                                    "title": {"type": ["null", "string"]},
                                },
                                "type": ["null", "object"],
                            },
                            "_ab_source_file_last_modified": {"type": "string"},
                            "_ab_source_file_url": {"type": "string"},
                        },
                    },
                    "name": "stream1",
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                }
            ]
        }
    )
).build()

avro_all_types_scenario = (
    TestScenarioBuilder()
    .set_name("avro_all_types_stream")
    .set_config(
        {
            "streams": [
                {
                    "name": "stream1",
                    "file_type": "avro",
                    "globs": ["*"],
                    "validation_policy": "emit_record",
                }
            ]
        }
    )
    .set_stream_reader(TemporaryAvroFilesStreamReader(files=_avro_all_types_file, file_type="avro"))
    .set_file_type("avro")
    .set_expected_records(
        [
            {
                "data": {
                    "col_bool": True,
                    "col_int": 27,
                    "col_long": 1992,
                    "col_float": 999.09723456,
                    "col_double": "9123456.12394",
                    "col_bytes": "\x00\x01\x02\x03",
                    "col_string": "Love It If We Made It",
                    "col_record": {"artist": "The 1975", "song": "About You", "year": 2022},
                    "col_enum": "POP_ROCK",
                    "col_array": [
                        "The 1975",
                        "I Like It When You Sleep, for You Are So Beautiful yet So Unaware of It",
                        "The 1975 A Brief Inquiry into Online Relationships",
                        "Notes on a Conditional Form",
                        "Being Funny in a Foreign Language",
                    ],
                    "col_map": {
                        "lead_singer": "Matty Healy",
                        "lead_guitar": "Adam Hann",
                        "bass_guitar": "Ross MacDonald",
                        "drummer": "George Daniel",
                    },
                    "col_fixed": "\x12\x34\x56\x78",
                    "col_decimal": 1234.56789,
                    "col_uuid": "123e4567-e89b-12d3-a456-426655440000",
                    "col_date": "2022-05-29",
                    "col_time_millis": "06:00:00.456000",
                    "col_time_micros": "12:00:00.456789",
                    "col_timestamp_millis": "2022-05-29T00:00:00.456000+00:00",
                    "col_timestamp_micros": "2022-05-30T00:00:00.456789+00:00",
                    "col_local_timestamp_millis": "2022-05-29T00:00:00.456000",
                    "col_local_timestamp_micros": "2022-05-30T00:00:00.456789",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a.avro",
                },
                "stream": "stream1",
            },
        ]
    )
    .set_expected_catalog(
        {
            "streams": [
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col_array": {"items": {"type": ["null", "string"]}, "type": ["null", "array"]},
                            "col_bool": {"type": ["null", "boolean"]},
                            "col_bytes": {"type": ["null", "string"]},
                            "col_double": {"type": ["null", "string"]},
                            "col_enum": {"enum": ["POP_ROCK", "INDIE_ROCK", "ALTERNATIVE_ROCK"], "type": ["null", "string"]},
                            "col_fixed": {"pattern": "^[0-9A-Fa-f]{8}$", "type": ["null", "string"]},
                            "col_float": {"type": ["null", "number"]},
                            "col_int": {"type": ["null", "integer"]},
                            "col_long": {"type": ["null", "integer"]},
                            "col_map": {"additionalProperties": {"type": ["null", "string"]}, "type": ["null", "object"]},
                            "col_record": {
                                "properties": {"artist": {"type": ["null", "string"]}, "song": {"type": ["null", "string"]}, "year": {"type": ["null", "integer"]}},
                                "type": ["null", "object"],
                            },
                            "col_string": {"type": ["null", "string"]},
                            "col_decimal": {"pattern": "^-?\\d{(1, 5)}(?:\\.\\d(1, 5))?$", "type": ["null", "string"]},
                            "col_uuid": {"type": ["null", "string"]},
                            "col_date": {"format": "date", "type": ["null", "string"]},
                            "col_time_millis": {"type": ["null", "integer"]},
                            "col_time_micros": {"type": ["null", "integer"]},
                            "col_timestamp_millis": {"format": "date-time", "type": ["null", "string"]},
                            "col_timestamp_micros": {"type": ["null", "string"]},
                            "_ab_source_file_last_modified": {"type": "string"},
                            "_ab_source_file_url": {"type": "string"},
                        },
                    },
                    "name": "stream1",
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                }
            ]
        }
    )
).build()

multiple_streams_avro_scenario = (
    TestScenarioBuilder()
    .set_name("multiple_streams_avro_stream")
    .set_config(
        {
            "streams": [
                {
                    "name": "songs_stream",
                    "file_type": "avro",
                    "globs": ["*_songs.avro"],
                    "validation_policy": "emit_record",
                },
                {
                    "name": "festivals_stream",
                    "file_type": "avro",
                    "globs": ["*_festivals.avro"],
                    "validation_policy": "emit_record",
                },
            ]
        }
    )
    .set_stream_reader(TemporaryAvroFilesStreamReader(files=_multiple_avro_stream_file, file_type="avro"))
    .set_file_type("avro")
    .set_expected_records(
        [
            {
                "data": {
                    "col_title": "Late Night",
                    "col_album": "A_MOMENT_APART",
                    "col_year": 2017,
                    "col_vocals": False,
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "odesza_songs.avro",
                },
                "stream": "songs_stream",
            },
            {
                "data": {
                    "col_title": "White Lies",
                    "col_album": "IN_RETURN",
                    "col_year": 2014,
                    "col_vocals": True,
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "odesza_songs.avro",
                },
                "stream": "songs_stream",
            },
            {
                "data": {
                    "col_title": "Wide Awake",
                    "col_album": "THE_LAST_GOODBYE",
                    "col_year": 2022,
                    "col_vocals": True,
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "odesza_songs.avro",
                },
                "stream": "songs_stream",
            },
            {
                "data": {
                    "col_title": "Sun Models",
                    "col_album": "SUMMERS_GONE",
                    "col_year": 2012,
                    "col_vocals": True,
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "odesza_songs.avro",
                },
                "stream": "songs_stream",
            },
            {
                "data": {
                    "col_title": "All We Need",
                    "col_album": "IN_RETURN",
                    "col_year": 2014,
                    "col_vocals": True,
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "odesza_songs.avro",
                },
                "stream": "songs_stream",
            },
            {
                "data": {
                    "col_name": "Coachella",
                    "col_location": {"country": "USA", "state": "California", "city": "Indio"},
                    "col_attendance": 250000,
                    "_ab_source_file_last_modified": "2023-06-06T03:54:07.000000Z",
                    "_ab_source_file_url": "california_festivals.avro",
                },
                "stream": "festivals_stream",
            },
            {
                "data": {
                    "col_name": "CRSSD",
                    "col_location": {"country": "USA", "state": "California", "city": "San Diego"},
                    "col_attendance": 30000,
                    "_ab_source_file_last_modified": "2023-06-06T03:54:07.000000Z",
                    "_ab_source_file_url": "california_festivals.avro",
                },
                "stream": "festivals_stream",
            },
            {
                "data": {
                    "col_name": "Lightning in a Bottle",
                    "col_location": {"country": "USA", "state": "California", "city": "Buena Vista Lake"},
                    "col_attendance": 18000,
                    "_ab_source_file_last_modified": "2023-06-06T03:54:07.000000Z",
                    "_ab_source_file_url": "california_festivals.avro",
                },
                "stream": "festivals_stream",
            },
            {
                "data": {
                    "col_name": "Outside Lands",
                    "col_location": {"country": "USA", "state": "California", "city": "San Francisco"},
                    "col_attendance": 220000,
                    "_ab_source_file_last_modified": "2023-06-06T03:54:07.000000Z",
                    "_ab_source_file_url": "california_festivals.avro",
                },
                "stream": "festivals_stream",
            },
        ]
    )
    .set_expected_catalog(
        {
            "streams": [
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col_title": {"type": ["null", "string"]},
                            "col_album": {"type": ["null", "string"], "enum": ["SUMMERS_GONE", "IN_RETURN", "A_MOMENT_APART", "THE_LAST_GOODBYE"]},
                            "col_year": {"type": ["null", "integer"]},
                            "col_vocals": {"type": ["null", "boolean"]},
                            "_ab_source_file_last_modified": {"type": "string"},
                            "_ab_source_file_url": {"type": "string"},
                        },
                    },
                    "name": "songs_stream",
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                },
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col_name": {"type": ["null", "string"]},
                            "col_location": {
                                "properties": {"country": {"type": ["null", "string"]}, "state": {"type": ["null", "string"]}, "city": {"type": ["null", "string"]}},
                                "type": ["null", "object"],
                            },
                            "col_attendance": {"type": ["null", "integer"]},
                            "_ab_source_file_last_modified": {"type": "string"},
                            "_ab_source_file_url": {"type": "string"},
                        },
                    },
                    "name": "festivals_stream",
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                },
            ]
        }
    )
).build()

avro_file_with_decimal_as_float_scenario = (
    TestScenarioBuilder()
    .set_name("avro_file_with_decimal_as_float_stream")
    .set_config(
        {
            "streams": [
                {
                    "name": "stream1",
                    "file_type": "avro",
                    "globs": ["*"],
                    "validation_policy": "emit_record",
                    "format": {
                        "avro": {
                            "filetype": "avro",
                            "decimal_as_float": True
                        }
                    }
                }
            ]
        }
    )
    .set_stream_reader(TemporaryAvroFilesStreamReader(files=_multiple_avro_combine_schema_file, file_type="avro"))
    .set_file_type("avro")
    .set_expected_records(
        [
            {
                "data": {
                    "col_double": 20.02,
                    "col_string": "Robbers",
                    "col_album": {"album": "The 1975"},
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a.avro",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col_double": 20.23,
                    "col_string": "Somebody Else",
                    "col_album": {"album": "I Like It When You Sleep, for You Are So Beautiful yet So Unaware of It"},
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a.avro",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col_double": 1975.1975,
                    "col_string": "It's Not Living (If It's Not with You)",
                    "col_song": {"title": "Love It If We Made It"},
                    "_ab_source_file_last_modified": "2023-06-06T03:54:07.000000Z",
                    "_ab_source_file_url": "b.avro",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col_double": 5791.5791,
                    "col_string": "The 1975",
                    "col_song": {"title": "About You"},
                    "_ab_source_file_last_modified": "2023-06-06T03:54:07.000000Z",
                    "_ab_source_file_url": "b.avro",
                },
                "stream": "stream1",
            },
        ]
    )
    .set_expected_catalog(
        {
            "streams": [
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col_double": {"type": ["null", "number"]},
                            "col_string": {"type": ["null", "string"]},
                            "col_album": {
                                "properties": {
                                    "album": {"type": ["null", "string"]},
                                },
                                "type": ["null", "object"],
                            },
                            "col_song": {
                                "properties": {
                                    "title": {"type": ["null", "string"]},
                                },
                                "type": ["null", "object"],
                            },
                            "_ab_source_file_last_modified": {"type": "string"},
                            "_ab_source_file_url": {"type": "string"},
                        },
                    },
                    "name": "stream1",
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                }
            ]
        }
    )
).build()



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import json
import logging
from copy import deepcopy
from json import JSONDecodeError
from typing import Any, Dict, Iterable, Iterator, List, Mapping, Optional, Union
from urllib.parse import parse_qs, urlparse

from airbyte_cdk.connector_builder.models import (
    AuxiliaryRequest,
    HttpRequest,
    HttpResponse,
    LogMessage,
    StreamRead,
    StreamReadPages,
    StreamReadSlices,
)
from airbyte_cdk.entrypoint import AirbyteEntrypoint
from airbyte_cdk.sources import AbstractSource
from airbyte_cdk.sources.declarative.declarative_source import DeclarativeSource
from airbyte_cdk.sources.utils.types import JsonType
from airbyte_cdk.utils import AirbyteTracedException
from airbyte_cdk.utils.datetime_format_inferrer import DatetimeFormatInferrer
from airbyte_cdk.utils.schema_inferrer import SchemaInferrer
from airbyte_protocol.models.airbyte_protocol import (
    AirbyteControlMessage,
    AirbyteLogMessage,
    AirbyteMessage,
    AirbyteTraceMessage,
    ConfiguredAirbyteCatalog,
    OrchestratorType,
    TraceType,
)
from airbyte_protocol.models.airbyte_protocol import Type as MessageType


class MessageGrouper:
    logger = logging.getLogger("airbyte.connector-builder")

    def __init__(self, max_pages_per_slice: int, max_slices: int, max_record_limit: int = 1000):
        self._max_pages_per_slice = max_pages_per_slice
        self._max_slices = max_slices
        self._max_record_limit = max_record_limit

    def get_message_groups(
        self,
        source: DeclarativeSource,
        config: Mapping[str, Any],
        configured_catalog: ConfiguredAirbyteCatalog,
        record_limit: Optional[int] = None,
    ) -> StreamRead:
        if record_limit is not None and not (1 <= record_limit <= 1000):
            raise ValueError(f"Record limit must be between 1 and 1000. Got {record_limit}")
        schema_inferrer = SchemaInferrer()
        datetime_format_inferrer = DatetimeFormatInferrer()

        if record_limit is None:
            record_limit = self._max_record_limit
        else:
            record_limit = min(record_limit, self._max_record_limit)

        slices = []
        log_messages = []
        latest_config_update: AirbyteControlMessage = None
        auxiliary_requests = []
        for message_group in self._get_message_groups(
                self._read_stream(source, config, configured_catalog),
                schema_inferrer,
                datetime_format_inferrer,
                record_limit,
        ):
            if isinstance(message_group, AirbyteLogMessage):
                log_messages.append(LogMessage(**{"message": message_group.message, "level": message_group.level.value}))
            elif isinstance(message_group, AirbyteTraceMessage):
                if message_group.type == TraceType.ERROR:
                    error_message = f"{message_group.error.message} - {message_group.error.stack_trace}"
                    log_messages.append(LogMessage(**{"message": error_message, "level": "ERROR"}))
            elif isinstance(message_group, AirbyteControlMessage):
                if not latest_config_update or latest_config_update.emitted_at <= message_group.emitted_at:
                    latest_config_update = message_group
            elif isinstance(message_group, AuxiliaryRequest):
                auxiliary_requests.append(message_group)
            elif isinstance(message_group, StreamReadSlices):
                slices.append(message_group)
            else:
                raise ValueError(f"Unknown message group type: {type(message_group)}")

        return StreamRead(
            logs=log_messages,
            slices=slices,
            test_read_limit_reached=self._has_reached_limit(slices),
            auxiliary_requests=auxiliary_requests,
            inferred_schema=schema_inferrer.get_stream_schema(
                configured_catalog.streams[0].stream.name
            ),  # The connector builder currently only supports reading from a single stream at a time
            latest_config_update=self._clean_config(latest_config_update.connectorConfig.config) if latest_config_update else None,
            inferred_datetime_formats=datetime_format_inferrer.get_inferred_datetime_formats(),
        )

    def _get_message_groups(
            self, messages: Iterator[AirbyteMessage], schema_inferrer: SchemaInferrer, datetime_format_inferrer: DatetimeFormatInferrer, limit: int
    ) -> Iterable[Union[StreamReadPages, AirbyteControlMessage, AirbyteLogMessage, AirbyteTraceMessage, AuxiliaryRequest]]:
        """
        Message groups are partitioned according to when request log messages are received. Subsequent response log messages
        and record messages belong to the prior request log message and when we encounter another request, append the latest
        message group, until <limit> records have been read.

        Messages received from the CDK read operation will always arrive in the following order:
        {type: LOG, log: {message: "request: ..."}}
        {type: LOG, log: {message: "response: ..."}}
        ... 0 or more record messages
        {type: RECORD, record: {data: ...}}
        {type: RECORD, record: {data: ...}}
        Repeats for each request/response made

        Note: The exception is that normal log messages can be received at any time which are not incorporated into grouping
        """
        records_count = 0
        at_least_one_page_in_group = False
        current_page_records: List[Mapping[str, Any]] = []
        current_slice_descriptor: Optional[Dict[str, Any]] = None
        current_slice_pages: List[StreamReadPages] = []
        current_page_request: Optional[HttpRequest] = None
        current_page_response: Optional[HttpResponse] = None

        while records_count < limit and (message := next(messages, None)):
            json_object = self._parse_json(message.log) if message.type == MessageType.LOG else None
            if json_object is not None and not isinstance(json_object, dict):
                raise ValueError(f"Expected log message to be a dict, got {json_object} of type {type(json_object)}")
            json_message: Optional[Dict[str, JsonType]] = json_object
            if self._need_to_close_page(at_least_one_page_in_group, message, json_message):
                self._close_page(current_page_request, current_page_response, current_slice_pages, current_page_records)
                current_page_request = None
                current_page_response = None

            if at_least_one_page_in_group and message.type == MessageType.LOG and message.log.message.startswith(AbstractSource.SLICE_LOG_PREFIX):
                yield StreamReadSlices(pages=current_slice_pages, slice_descriptor=current_slice_descriptor)
                current_slice_descriptor = self._parse_slice_description(message.log.message)
                current_slice_pages = []
                at_least_one_page_in_group = False
            elif message.type == MessageType.LOG and message.log.message.startswith(AbstractSource.SLICE_LOG_PREFIX):
                # parsing the first slice
                current_slice_descriptor = self._parse_slice_description(message.log.message)
            elif message.type == MessageType.LOG:
                if json_message is not None and self._is_http_log(json_message):
                    if self._is_auxiliary_http_request(json_message):
                        airbyte_cdk = json_message.get("airbyte_cdk", {})
                        if not isinstance(airbyte_cdk, dict):
                            raise ValueError(f"Expected airbyte_cdk to be a dict, got {airbyte_cdk} of type {type(airbyte_cdk)}")
                        stream = airbyte_cdk.get("stream", {})
                        if not isinstance(stream, dict):
                            raise ValueError(f"Expected stream to be a dict, got {stream} of type {type(stream)}")
                        title_prefix = (
                           "Parent stream: " if stream.get("is_substream", False) else ""
                        )
                        http = json_message.get("http", {})
                        if not isinstance(http, dict):
                            raise ValueError(f"Expected http to be a dict, got {http} of type {type(http)}")
                        yield AuxiliaryRequest(
                            title=title_prefix + str(http.get("title", None)),
                            description=str(http.get("description", None)),
                            request=self._create_request_from_log_message(json_message),
                            response=self._create_response_from_log_message(json_message),
                        )
                    else:
                        at_least_one_page_in_group = True
                        current_page_request = self._create_request_from_log_message(json_message)
                        current_page_response = self._create_response_from_log_message(json_message)
                else:
                    yield message.log
            elif message.type == MessageType.TRACE:
                if message.trace.type == TraceType.ERROR:
                    yield message.trace
            elif message.type == MessageType.RECORD:
                current_page_records.append(message.record.data)
                records_count += 1
                schema_inferrer.accumulate(message.record)
                datetime_format_inferrer.accumulate(message.record)
            elif message.type == MessageType.CONTROL and message.control.type == OrchestratorType.CONNECTOR_CONFIG:
                yield message.control
        else:
            if current_page_request or current_page_response or current_page_records:
                self._close_page(current_page_request, current_page_response, current_slice_pages, current_page_records)
                yield StreamReadSlices(pages=current_slice_pages, slice_descriptor=current_slice_descriptor)

    @staticmethod
    def _need_to_close_page(at_least_one_page_in_group: bool, message: AirbyteMessage, json_message: Optional[Dict[str, Any]]) -> bool:
        return (
            at_least_one_page_in_group
            and message.type == MessageType.LOG
            and (MessageGrouper._is_page_http_request(json_message) or message.log.message.startswith("slice:"))
        )

    @staticmethod
    def _is_page_http_request(json_message: Optional[Dict[str, Any]]) -> bool:
        if not json_message:
            return False
        else:
            return MessageGrouper._is_http_log(json_message) and not MessageGrouper._is_auxiliary_http_request(json_message)

    @staticmethod
    def _is_http_log(message: Dict[str, JsonType]) -> bool:
        return bool(message.get("http", False))

    @staticmethod
    def _is_auxiliary_http_request(message: Optional[Dict[str, Any]]) -> bool:
        """
        A auxiliary request is a request that is performed and will not directly lead to record for the specific stream it is being queried.
        A couple of examples are:
        * OAuth authentication
        * Substream slice generation
        """
        if not message:
            return False

        is_http = MessageGrouper._is_http_log(message)
        return is_http and message.get("http", {}).get("is_auxiliary", False)

    @staticmethod
    def _close_page(current_page_request: Optional[HttpRequest], current_page_response: Optional[HttpResponse], current_slice_pages: List[StreamReadPages], current_page_records: List[Mapping[str, Any]]) -> None:
        """
        Close a page when parsing message groups
        """
        current_slice_pages.append(
            StreamReadPages(request=current_page_request, response=current_page_response, records=deepcopy(current_page_records))  # type: ignore
        )
        current_page_records.clear()

    def _read_stream(self, source: DeclarativeSource, config: Mapping[str, Any], configured_catalog: ConfiguredAirbyteCatalog) -> Iterator[AirbyteMessage]:
        # the generator can raise an exception
        # iterate over the generated messages. if next raise an exception, catch it and yield it as an AirbyteLogMessage
        try:
            yield from AirbyteEntrypoint(source).read(source.spec(self.logger), config, configured_catalog, {})
        except Exception as e:
            error_message = f"{e.args[0] if len(e.args) > 0 else str(e)}"
            yield AirbyteTracedException.from_exception(e, message=error_message).as_airbyte_message()

    @staticmethod
    def _parse_json(log_message: AirbyteLogMessage) -> JsonType:
        # TODO: As a temporary stopgap, the CDK emits request/response data as a log message string. Ideally this should come in the
        # form of a custom message object defined in the Airbyte protocol, but this unblocks us in the immediate while the
        # protocol change is worked on.
        try:
            json_object: JsonType = json.loads(log_message.message)
            return json_object
        except JSONDecodeError:
            return None

    @staticmethod
    def _create_request_from_log_message(json_http_message: Dict[str, Any]) -> HttpRequest:
        url = urlparse(json_http_message.get("url", {}).get("full", ""))
        full_path = f"{url.scheme}://{url.hostname}{url.path}" if url else ""
        request = json_http_message.get("http", {}).get("request", {})
        parameters = parse_qs(url.query) or None
        return HttpRequest(
            url=full_path,
            http_method=request.get("method", ""),
            headers=request.get("headers"),
            parameters=parameters,
            body=request.get("body", {}).get("content", ""),
        )

    @staticmethod
    def _create_response_from_log_message(json_http_message: Dict[str, Any]) -> HttpResponse:
        response = json_http_message.get("http", {}).get("response", {})
        body = response.get("body", {}).get("content", "")
        return HttpResponse(status=response.get("status_code"), body=body, headers=response.get("headers"))

    def _has_reached_limit(self, slices: List[StreamReadSlices]) -> bool:
        if len(slices) >= self._max_slices:
            return True

        record_count = 0

        for _slice in slices:
            if len(_slice.pages) >= self._max_pages_per_slice:
                return True
            for page in _slice.pages:
                record_count += len(page.records)
                if record_count >= self._max_record_limit:
                    return True
        return False

    def _parse_slice_description(self, log_message: str) -> Dict[str, Any]:
        return json.loads(log_message.replace(AbstractSource.SLICE_LOG_PREFIX, "", 1))  # type: ignore

    @staticmethod
    def _clean_config(config: Dict[str, Any]) -> Dict[str, Any]:
        cleaned_config = deepcopy(config)
        for key in config.keys():
            if key.startswith("__"):
                del cleaned_config[key]
        return cleaned_config



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from typing import Type

import pytest
from airbyte_cdk.sources.file_based.config.file_based_stream_config import AvroFormat, CsvFormat, ParquetFormat
from jsonschema import ValidationError, validate
from pydantic import BaseModel


@pytest.mark.parametrize(
    "file_format, file_type, expected_error",
    [
        pytest.param(ParquetFormat, "parquet", None, id="test_parquet_format_is_a_valid_parquet_file_type"),
        pytest.param(AvroFormat, "avro", None, id="test_avro_format_is_a_valid_avro_file_type"),
        pytest.param(CsvFormat, "parquet", ValidationError, id="test_csv_format_is_not_a_valid_parquet_file_type"),
    ]
)
def test_parquet_file_type_is_not_a_valid_csv_file_type(file_format: BaseModel, file_type: str, expected_error: Type[Exception]) -> None:
    format_config = {
        file_type: {
            "filetype": file_type,
            "decimal_as_float": True
        }
    }

    if expected_error:
        with pytest.raises(expected_error):
            validate(instance=format_config[file_type], schema=file_format.schema())
    else:
        validate(instance=format_config[file_type], schema=file_format.schema())
